/* Copyright (c) 2009, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <poll.h>
#include <unistd.h>
#include <sys/mman.h>	/* for PAGE_SIZE */

/* how often (in miliseconds) an update is performed */
#define UPDATE_INTERVAL		3000

static char *top_printtime(unsigned int time)
{
	static char buf[10];	/* see printf below for length explanation */

	int days, hours, minutes, seconds, msecs;

	msecs = time % 1000;
	time /= 1000;
	seconds = time % 60;
	time /= 60;
	minutes = time % 60;
	time /= 60;
	hours = time % 24;
	time /= 24;
	days = time;

	if (days)
		sprintf(buf, "%2id %2i:%02i", days, hours, minutes);
	else if (hours)
		sprintf(buf, " %2i:%02i:%02i", hours, minutes, seconds);
	else
		sprintf(buf, "%2i:%02i.%03i", minutes, seconds, msecs);

	return buf;
}

static int top_cpuusage(unsigned int newct, unsigned int oldct)
{
	if (oldct >= newct || (newct - oldct) > UPDATE_INTERVAL)
		/* no / invalid data or just 0 (in case of oldct == newct) */
		return 0;

	return 100 * (newct - oldct) / UPDATE_INTERVAL;
}

static void top_print(void)
{
	int i;
	int memused;
	int ptotal, pzombie, psleeping, prunning, pstopped;
	int firstrun = 0;

	struct sched_info si;
	static struct sched_procinfo *spi;
	static unsigned int *cputimes;

	sched_info(&si);

	if (!spi) {
		/* si.procs does not change, so this is save */
		spi = malloc(sizeof(struct sched_procinfo) * si.procs);
		cputimes = malloc(sizeof(unsigned int) * si.procs);

		firstrun = 1;
		printf("collecting data for %i miliseconds...\n", UPDATE_INTERVAL);
	}

	ptotal = sched_list(spi, si.procs);

	if (firstrun) {
		/* copy cpu times, normally done after printing */
		for (i = 0; i < ptotal; i++)
			cputimes[i] = spi[i].cputime;

		/* cannot print anything on first run (no CPU percentages
		 * available) */
		return;
	}

	/* build up header information */
	memused = pzombie = psleeping = prunning = pstopped = 0;
	for (i = 0; i < ptotal; i++) {
		switch (spi[i].state) {
			case STATE_ZOMBIE: pzombie++; break;
			case STATE_RUNNING: prunning++; break;
			case STATE_SLEEPING: psleeping++; break;
			case STATE_STOPPED: pstopped++; break;
		}

		memused += spi[i].memsz;
	}

	/* erase screen */
	write(STDOUT_FILENO, "\e[H\e[J", 6);

	/* print header information */
	printf("scheduling frequency: %i Hz, maximal processes: %i\n",
	       si.hz, si.procs);
	printf("processes:  total %i, running %i, sleeping %i, zombie %i, stopped %i\n",
	       ptotal, prunning, psleeping, pzombie, pstopped);
	printf("load average:  %i.%03i %i.%03i %i.%03i\n",
	       si.la1 / 1000, si.la1 % 1000,
	       si.la5 / 1000, si.la5 % 1000,
	       si.la15 / 1000, si.la15 % 1000);
	printf("memory:  total %i kB, used %i kB, free %i kB\n",
	       (si.memtotal * PAGE_SIZE) >> 10, (memused * PAGE_SIZE) >> 10,
	       ((si.memtotal - memused) * PAGE_SIZE) >> 10);
	printf("uptime:  %s\n", top_printtime(si.uptime));
	printf("cpu usage:  running %3i %%, sleeping %3i %%\n",
	       100 - top_cpuusage(spi[0].cputime, cputimes[0]),
	       top_cpuusage(spi[0].cputime, cputimes[0]));
	printf("\n");

	/* print process table headline */
	printf(" pid ppid s memsz  cpu-time cu%% argv\n");

	/* TODO: sort them by cpu usage */

	/* print process table, first running processes, then the others */
	for (i = 0; i < ptotal; i++)
		if (spi[i].state == STATE_RUNNING)
			printf("%4i %4i %c %5i %s %3i %s\n",
			       spi[i].pid, spi[i].ppid, spi[i].state,
			       (spi[i].memsz * PAGE_SIZE) >> 10,
			       top_printtime(spi[i].cputime),
			       top_cpuusage(spi[i].cputime, cputimes[i]),
			       spi[i].cmd);
	for (i = 0; i < ptotal; i++)
		if (spi[i].state != STATE_RUNNING)
			printf("%4i %4i %c %5i %s %3i %s\n",
			       spi[i].pid, spi[i].ppid, spi[i].state,
			       (spi[i].memsz * PAGE_SIZE) >> 10,
			       top_printtime(spi[i].cputime),
			       top_cpuusage(spi[i].cputime, cputimes[i]),
			       spi[i].cmd);

	/* copy CPU times for CPU usage calculation */
	for (i = 0; i < ptotal; i++)
		cputimes[i] = spi[i].cputime;
}

int main(int argc, char **argv)
{
	(void) argc;
	(void) argv;

	while (1) {
		/* print info */
		top_print();

		/* wait */
		if (poll(0, 0, UPDATE_INTERVAL) == -1)
			return EXIT_FAILURE;
		/* if timed out (r = 0), just update */
	}
}
