/* Copyright (c) 2009-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#pragma once

#include <errno.h>
#include <sys/kernel/syscall.h>

/* compiler attributes */
#define __noreturn	__attribute__ ((noreturn)) void
#define __pure		__attribute__ ((pure))

/* This is not to be used in the userspace.  Though it is defined as some libc
 * functions are included in the kernelspace and there it is used. Therefore,
 * __optimize in the userspace does not what it does in kernelspace. */
#define __optimize

/* syscall makros setting errno. Can only work for functions which shall return
 * -1 on error. (no mmap(), ...) */

/* used in syscall makros */
#define __errno_stub		\
	if (r < 0) {		\
		errno = -r;	\
		return -1;	\
	}			\
	return r

/* syscall returning always int 0 (no errno) with no parameter */
#define __syscalle0(ret, name, sc)			\
	static inline ret name(void)			\
	{						\
		asm volatile ("swi %[s]" ::		\
			      [s] "i" (sc));		\
		return 0;				\
	}

/* syscall returning int with no parameter */
#define __syscalli0(ret, name, sc)			\
	static inline ret name(void)			\
	{						\
		register int r asm("r0");		\
		asm volatile ("swi %[s]" :		\
			      "=r" (r) :		\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 1 parameter */
#define __syscalli1(ret, name, sc, at)			\
	static inline ret name(at a)			\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r) :		\
			      "r" (a1),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 1 parameter which is an output memory region */
#define __syscalli1o1(ret, name, sc, at)		\
	static inline ret name(at a)			\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r),			\
			      "=m" (*a) :		\
			      "r" (a1),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 2 parameters */
#define __syscalli2(ret, name, sc, at, bt)		\
	static inline ret name(at a, bt b)		\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r) :		\
			      "r" (a1),			\
			      "r" (a2),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 2 parameters where first parameter is an input
 * memory region. */
#define __syscalli2i1(ret, name, sc, at, bt)		\
	static inline ret name(at a, bt b)		\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r) :		\
			      "r" (a1),			\
			      "m" (*a),			\
			      "r" (a2),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 2 parameters where first parameter is an output
 * memory region. */
#define __syscalli2o1(ret, name, sc, at, bt)		\
	static inline ret name(at a, bt b)		\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r),			\
			      "=m" (*a) :		\
			      "r" (a1),			\
			      "r" (a2),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 3 parameters */
#define __syscalli3(ret, name, sc, at, bt, ct)		\
	static inline ret name(at a, bt b, ct c)	\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		register int a3 asm("r2") = (int) c;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r) :		\
			      "r" (a1),			\
			      "r" (a2),			\
			      "r" (a3),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 3 parameters where first parameter is an
 * input/output memory region */
#define __syscalli3i1o1(ret, name, sc, at, bt, ct)	\
	static inline ret name(at a, bt b, ct c)	\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		register int a3 asm("r2") = (int) c;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r),			\
			      "=m" (*a) :		\
			      "r" (a1),			\
			      "m" (*a),			\
			      "r" (a2),			\
			      "r" (a3),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 3 parameters where second parameter is an input
 * memory region and third parameter an output memory region. */
#define __syscalli3i2o3(ret, name, sc, at, bt, ct)	\
	static inline ret name(at a, bt b, ct c)	\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		register int a3 asm("r2") = (int) c;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r),			\
			      "=m" (*c) :		\
			      "r" (a1),			\
			      "r" (a2),			\
			      "m" (*b),			\
			      "r" (a3),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 3 parameters where second parameter is an output
 * memory region */
#define __syscalli3o2(ret, name, sc, at, bt, ct)	\
	static inline ret name(at a, bt b, ct c)	\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		register int a3 asm("r2") = (int) c;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r),			\
			      "=m" (*b) :		\
			      "r" (a1),			\
			      "r" (a2),			\
			      "r" (a3),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 3 parameters where second parameter is an input
 * memory region */
#define __syscalli3i2(ret, name, sc, at, bt, ct)	\
	static inline ret name(at a, bt b, ct c)	\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		register int a3 asm("r2") = (int) c;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r) :		\
			      "r" (a1),			\
			      "r" (a2),			\
			      "m" (*b),			\
			      "r" (a3),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}

/* syscall returning int with 3 parameters where all parameters are input memory
 * regions. */
#define __syscalli3i123(ret, name, sc, at, bt, ct)	\
	static inline ret name(at a, bt b, ct c)	\
	{						\
		register int r asm("r0");		\
		register int a1 asm("r0") = (int) a;	\
		register int a2 asm("r1") = (int) b;	\
		register int a3 asm("r2") = (int) c;	\
		asm volatile ("swi %[s]" :		\
			      "=r" (r) :		\
			      "r" (a1),			\
			      "m" (*a),			\
			      "r" (a2),			\
			      "m" (*b),			\
			      "r" (a3),			\
			      "m" (*c),			\
			      [s] "i" (sc));		\
		__errno_stub;				\
	}
