/* Copyright (c) 2008-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#ifndef KERNEL
#include <errno.h>
#include <stdio.h>
#include <string.h>
#endif

/* makros for printf */
#define CHAR_CPY(c)				\
	dest[rv++] = c;
#define FILL_CPY(x)				\
	for (i = minlen - (x); i > 0; i--)	\
		CHAR_CPY(fillc)
#define STR_CPY(s)				\
	i = 0;					\
	while (s[i])				\
		CHAR_CPY(s[i++]);

static void itos(unsigned long id, char *s, int base, int is_signed)
{
	static const char chars[] = "0123456789abcdef";
	unsigned char j = 0;
	int negative = 0;
	char *p1, *p2, tmp;

	if (is_signed && (signed) id < 0) {
		negative = 1;
		id = -(signed) id;
	}

	do {
		s[j++] = chars[id % base];
		id /= base;
	} while (id != 0);

	if (negative)
		s[j++] = '-';

	s[j] = 0;

	/* reverse the buf (we have written in wrong order) */
	p1 = s;
	p2 = &s[j - 1];
	while (p1 < p2) {
		tmp = *p1;
		*p1 = *p2;
		*p2 = tmp;
		p1++;
		p2--;
	}
}

int vsprintf(char *dest, const char *format, va_list ap)
{
	int rv = 0, i;
	char c;
	char buf[12];
	const char *s;

	int minlen, alt;
	char fillc;

	/* TODO: make this POSIX conform, smaller and vsnprintf */

	while ((c = *format++) != 0) {
		/* variable references begin with '%' */
		if (c != '%') {
			CHAR_CPY(c);
			continue;
		}

		/* reset to default config */
		minlen = 0;
		fillc = ' ';
		alt = 0;

		c = *format++;

		/* # specifies the "alternative format", e.g. 0x before
		 * hexadecimal numbers. */
		if (c == '#') {
			alt = 1;
			c = *format++;
		}

		/* may be followed by '0' to tell that they should be
		 * right-aligned by filling with '0' instead of ' ' */
		if (c == '0') {
			fillc = c;
			c = *format++;
		}

		/* may be followed by the minimum field size */
		while ('0' <= c && c <= '9') {
			minlen *= 10;
			minlen += c - '0';
			c = *format++;
		}

		/* must be followed by type indicator */
		switch (c) {
			case 'c':
				CHAR_CPY((char) va_arg(ap, int));
				break;
			case 'o':
				itos(va_arg(ap, unsigned), buf, 8, 0);
				if (alt) {
					if (fillc == ' ')
						FILL_CPY(strlen(buf) + 1);
					STR_CPY("0");
					if (fillc == '0')
						FILL_CPY(strlen(buf) + 1);
				} else
					FILL_CPY(strlen(buf));
				STR_CPY(buf);
				break;
			case 'd':
			case 'i':
				itos(va_arg(ap, int), buf, 10, 1);
				FILL_CPY(strlen(buf));
				STR_CPY(buf);
				break;
			case 'x':
				itos(va_arg(ap, unsigned), buf, 16, 0);
				if (alt) {
					if (fillc == ' ')
						FILL_CPY(strlen(buf) + 2);
					STR_CPY("0x");
					if (fillc == '0')
						FILL_CPY(strlen(buf) + 2);
				} else
					FILL_CPY(strlen(buf));
				STR_CPY(buf);
				break;
			case 's':
				s = va_arg(ap, const char *);
				if (!s)
					s = "(null)";
				FILL_CPY(strlen(s));
				STR_CPY(s);
				break;
			case '%':
				CHAR_CPY('%');
				break;
			default:
				/* we must abort here, because else there would
				 * be the possibility to read behind the string
				 * on fake situations like %#\0 */
#ifdef KERNEL
				return -EILSEQ;
#else
				errno = EILSEQ;
				return -1;
#endif
		}
	}

	dest[rv] = 0;

	return rv;
}
