/* Copyright (c) 2008-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#ifndef KERNEL
#include <stdlib.h>
#include <sys/mman.h>
#include <stdio.h>
#endif

#define BLOCKSIZE	4			/* useful, since structs are 32
						 * bit aligned and padded */
#define BPE		(PAGE_SIZE / BLOCKSIZE)	/* blocks per mpg entry */
#define BITMAPSIZE	(BPE / 32)		/* number of ints of bitmap */
#define FIRSTDATABLOCK	(BITMAPSIZE + 2)	/* 1st block useable for data */
#define HEADSIZE	(FIRSTDATABLOCK * 4)	/* size of mpg header */
#define DATASIZE	(PAGE_SIZE - HEADSIZE)	/* size of data area */
#define INITFREE	(BPE - FIRSTDATABLOCK)	/* free blocks on freshly
						 * allocated mpg */

/* structure of a page used by malloc routines */
struct mpg {
	struct mpg *next;
	int freeblocks;
	int bitmap[BITMAPSIZE];
	char data[DATASIZE];
};

#ifndef KERNEL
extern struct mpg *__malloc_pages;	/* head of linked list */
#else
static struct mpg *__malloc_pages;
#endif

static inline struct mpg *find_mpg(void *ptr)
{
	struct mpg *mpg;

	/* search belonging mpg */
	for (mpg = __malloc_pages; mpg; mpg = mpg->next)
		if ((unsigned long) mpg < (unsigned long) ptr &&
		    (unsigned long) ptr < (unsigned long) mpg + PAGE_SIZE)
			/* found */
			return mpg;

	/* not found.
	 * Does not check all cases of invalid free()s, but it does not lack
	 * performance.
	 * POSIX does not the define the behaviour of free() when attempting to
	 * free invalid memory portions */
#ifdef KERNEL
	kpanic("invalid_ptr: %#010x", (unsigned int) ptr);
#else
	fprintf(stderr, "\ninvalid ptr: %#010x\n", (unsigned int) ptr);
	abort();
#endif
}

#ifndef KERNEL	/* kernel has own blocks() function in lib.h */
static inline unsigned int blocks(int size, int blocksize)
{
	int r = size / blocksize;

	if (size % blocksize != 0)
		r++;

	return r;
}
#endif

static inline void do_free(struct mpg *mpg, int from, int to)
{
	int i;
	for (i = from; i < to; i++)
		mpg->bitmap[i / 32] &= ~(1 << (i % 32));
	mpg->freeblocks += to - from;
}

static inline void do_alloc(struct mpg *mpg, int from, int to)
{
	int i;
	for (i = from; i < to; i++)
		mpg->bitmap[i / 32] |= 1 << (i % 32);
	mpg->freeblocks -= to - from;
}
