/*-
 * Copyright (c) 2010 Alexander Graf.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef KERNEL
#include <string.h>
#include "malloc.h"
#endif

void *realloc(void *ptr, size_t size)
{
	void *newptr;
	struct mpg *mpg;
	int oldblkcnt, newblkcnt;
	int blknr;
	int i;

	if (!ptr)
		return malloc(size);

	if (!size) {
		free(ptr);
		return 0;
	}

	mpg = find_mpg(ptr);

	ptr -= 4;
	oldblkcnt = *(int *) ptr;

	size += 4;
	newblkcnt = blocks(size, BLOCKSIZE);

	if (oldblkcnt == newblkcnt)
		/* nothing need to be done */
		return ptr + 4;

	/* get block number in this mpg */
	blknr = ((unsigned long) ptr - (unsigned long) mpg) / BLOCKSIZE;

	if (oldblkcnt > newblkcnt) {
		/* Shrinking allocation.  We just free the spare blocks. */
		do_free(mpg, blknr + newblkcnt, blknr + oldblkcnt);
		*(int *) ptr = newblkcnt;
		return ptr + 4;
	}

	/* Enlarge allocation.  If enough blocks behind old allocation are free,
	 * allocate them.  Else, create new allocation, memcpy content and free
	 * old allocation. */

	if (mpg->freeblocks < newblkcnt - oldblkcnt)
		/* Too few free blocks in the mpg. */
		goto new_allocation;
	if (blknr + newblkcnt >= BPE)
		/* Too close to the end of the mpg. */
		goto new_allocation;

	for (i = blknr + oldblkcnt; i < blknr + newblkcnt; i++)
		if (mpg->bitmap[i / 32] & 1 << (i % 32))
			/* not enough blocks free behind allocation */
			goto new_allocation;

	do_alloc(mpg, blknr + oldblkcnt, blknr + newblkcnt);
	*(int *) ptr = newblkcnt;
	return ptr + 4;

new_allocation:
	newptr = malloc(size - 4);
	memcpy(newptr, ptr + 4, (oldblkcnt - blocks(4, BLOCKSIZE)) * BLOCKSIZE);
	free(ptr + 4);
	return newptr;
}
