#!/bin/sh

printmsg()
{
	echo -e "\e[31m>>> " $* "\e[0m"
}

collect_sysinfo()
{
	which wget >/dev/null 2>&1 && fetch_prog="wget"
	which fetch >/dev/null 2>&1 && fetch_prog="fetch"
	which curl >/dev/null 2>&1 && fetch_prog="curl"
	printmsg "program to fetch from http/ftp: $fetch_prog"

	which make >/dev/null 2>&1 && build_prog="make"
	which gmake >/dev/null 2>&1 && build_prog="gmake"
	printmsg "make implementation: $build_prog"

	[ -f /usr/local/lib/libmpfr.so ] && mpfr_dir="/usr/local"
	[ -f /usr/lib/libmpfr.so ] && mpfr_dir="/usr"
	printmsg "mpfr installation dir: $mpfr_dir"

	[ -f /usr/local/lib/libgmp.so ] && gmp_dir="/usr/local"
	[ -f /usr/lib/libgmp.so ] && gmp_dir="/usr"
	printmsg "gmp installation dir: $gmp_dir"

	[ -f /usr/local/lib/libmpc.so ] && mpc_dir="/usr/local"
	[ -f /usr/lib/libmpc.so ] && mpc_dir="/usr"
	printmsg "mpc installation dir: $mpc_dir"

	printmsg
	printmsg "If this is correct, press enter now. If not, press ^C and try"
	printmsg "to fix it."
	read
}

fetch_src()
{
	from_net=0

	(echo $1 | grep -qe "^http://") && from_net=1
	(echo $1 | grep -qe "^ftp://") && from_net=1

	if [ $from_net -eq 1 ]; then
		case $fetch_prog in
			wget)	wget $1 -O -;;
			fetch)	fetch $1 -o -;;
			curl)	curl $1;;
		esac
	else
		cat $1
	fi
}

extract_src()
{
	(echo $1 | grep -qe "bz2$") && alg="bunzip2"
	(echo $1 | grep -qe "gz$") && alg="gunzip"
	(echo $1 | grep -qe "lzma$") && alg="unlzma"

	[ -z $alg ] && alg="cat"

	$alg | tar -xf - -C $2
}

prepare_src()
{
	printmsg "Extracting $1 to $2..."
	fetch_src $1 | extract_src $1 $2
}

enable_disable()
{
	if [ $1 = "yes" ]; then
		echo "--enable-$2"
	else
		echo "--disable-$2"
	fi
}

configure_src()
{
	config_args="$* --target=$TARGET --prefix=$PREFIX --with-local-prefix=$PREFIX --disable-werror"

	config_args="$config_args `enable_disable $MULTILIB "multilib"`"
	config_args="$config_args `enable_disable $SHARED "shared"`"
	config_args="$config_args `enable_disable $NLS "nls"`"
	config_args="$config_args `enable_disable $SSP "libssp"`"

	if [ ! -z $DEFAULT_ARCH ]; then
		config_args="$config_args --with-arch=$DEFAULT_ARCH"
	fi

	if [ ! -z $DEFAULT_CPU ]; then
		config_args="$config_args --with-cpu=$DEFAULT_CPU"
	fi

	printmsg "configuring with: $config_args"
	$config_args || exit 1
}

build_src()
{
	printmsg "building"
	$build_prog -j$BUILD_THREADS || exit 1
}

install_src()
{
	printmsg "installing"
	$build_prog install || exit 1
}

cleanup_src()
{
	printmsg "removing $1"
	rm -rf $1
}

install_binutils()
{
	binutils_tmp=`mktemp -d $TMPDIR/binutils.XXXX`

	prepare_src $BINUTILS_SRC $binutils_tmp
	srcdir=$binutils_tmp/*
	mkdir $binutils_tmp/work
	cd $binutils_tmp/work
	configure_src $srcdir/configure $BINUTILS_EXTRA_CONF
	build_src
	install_src
	cd $OLDPWD
	cleanup_src $binutils_tmp
}

install_gcc()
{
	gcc_tmp=`mktemp -d $TMPDIR/gcc.XXXX`

	[ ! -z $mpfr_dir ] && libpaths="--with-mpfr=$mpfr_dir"
	[ ! -z $gmp_dir ] && libpaths="$libpaths --with-gmp=$gmp_dir"
	[ ! -z $mpc_dir ] && libpaths="$libpaths --with-mpc=$mpc_dir"

	prepare_src $GCC_SRC $gcc_tmp
	srcdir=$gcc_tmp/*
	mkdir $gcc_tmp/work
	cd $gcc_tmp/work
	configure_src $srcdir/configure $GCC_EXTRA_CONF	--enable-languages=$LANGUAGES $libpaths --without-headers
	build_src
	install_src
	cd $OLDPWD
	cleanup_src $gcc_tmp
}

collect_sysinfo
. ./mktoolchain.conf

export CFLAGS
export LDFLAGS

[ ! $SKIP_BINUTILS = "yes" ] && install_binutils
[ ! $SKIP_GCC = "yes" ] && install_gcc

printmsg "done"
