/* Copyright (c) 2009-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "bio.h"
#include "errno.h"
#include "lib.h"
#include "sched.h"
#include "syscall.h"

extern const struct io_handle devfs_ioh;
extern const struct io_handle minix_ioh;
extern const struct io_handle ext2fs_ioh;

static const struct {
	const char *name;
	const struct io_handle *ioh;
} fstypes[] = {
#ifdef CONFIG_EXT2FS
	{ "ext2fs", &ext2fs_ioh },
#endif
#ifdef CONFIG_MINIX
	{ "minix", &minix_ioh },
#endif
	{ 0, 0 }
};

static struct mountpoint *mntpts;

int def_open(struct filedes *fd, const char *path)
{
	(void) fd;
	(void) path;
	return 0;
}

int def_close(struct filedes *fd)
{
	(void) fd;
	return 0;
}

void def_clone_fd(struct filedes *dst, struct filedes *src)
{
	(void) dst;
	(void) src;
}

int def_poll(struct filedes *fd, int event)
{
	(void) fd;
	return (POLLRDNORM | POLLWRNORM) & event;
}

void bio_init(void)
{
	/* mount /dev filesystem */
	mntpts = malloc(sizeof(struct mountpoint));
	mntpts->next = 0;
	mntpts->ioh = &devfs_ioh;
	mntpts->path = "/dev/";
	mntpts->ioh->mount(mntpts);
}

int bio_mount(const char *dev, const char *path, const char *type)
{
	int r;
	struct mountpoint *m;
	struct mountpoint *mntp = malloc(sizeof(struct mountpoint));

	/* setup mntp struct */
	mntp->next = 0;
	mntp->path = path;
	mntp->ud   = 0;

	/* allocate mntp in linked mountpoint list */
	for (m = mntpts; m->next; m = m->next) ;
	m->next = mntp;

	/* search io_handle for filesystem type */
	for (r = 0; fstypes[r].name; r++)
		if (!strcmp(type, fstypes[r].name)) {
			mntp->ioh = fstypes[r].ioh;
			break;
		}
	if (!fstypes[r].name) {
		printf("bio: invalid fstype: %s\n", type);
		r = -EINVAL;
		goto fail;
	}

	/* open source device */
	r = bio_open(0, dev, O_RDWR, &mntp->src);
	if (r < 0 && r != -EROFS)
		goto fail;
	if (r < 0) {
		r = bio_open(0, dev, O_RDONLY, &mntp->src);
		if (r < 0)
			goto fail;
		printf("%s: opened read-only\n", dev);
	}

	/* call filesystem driver */
	r = mntp->ioh->mount(mntp);
	if (r < 0)
		goto fail;

	return 0;

fail:
	m->next = 0;
	free(mntp);
	return r;
}

/* TODO: replace identical stuff by makros / functions */

int bio_open(int tid, const char *path, int flags, struct filedes **fds)
{
	struct mountpoint *m, *mp = 0;
	struct filedes *f, *fd;
	unsigned int ms = 0;
	int poff = 0;
	int fdn = -1;
	int r;

	/* TODO: do not allow O_RDWR or O_WRONLY opens on files on a read-only
	 * filesystem */

	/* search mount of path */
	for (m = mntpts; m; m = m->next) {
		if (strlen(m->path) <= ms)
			continue;

		if (strlen(m->path) > strlen(path))
			continue;

		if (!strncmp(path, m->path, strlen(m->path))) {
			ms = strlen(m->path);
			poff = strlen(m->path);
			mp = m;
		}
	}
	if (!mp)
		return -ENOENT;

	/* search free fd number */
	/* TODO: bad algorithm */
	for (fd = procs[tid]->fds; fd; fd = fd->next)
		if (fdn == -1 || fdn < fd->fd)
			fdn = fd->fd;
	fdn++;

	/* create filedes struct */
	fd = malloc(sizeof(struct filedes));
	if (!procs[tid]->fds)
		procs[tid]->fds = fd;
	else {
		for (f = procs[tid]->fds; f->next; f = f->next) ;
		f->next = fd;
	}

	/* setup filedes struct */
	fd->next = 0;
	fd->owner = procs[tid];
	fd->mntp = mp;
	fd->ioh = mp->ioh;
	fd->fd = fdn;
	fd->flags = flags;
	fd->pos = 0;
	fd->ud = 0;

	/* call io handler */
	if ((r = fd->ioh->open(fd, &path[poff])) >= 0) {
		if (fds)
			*fds = fd;
		return fd->fd;
	}

	/* io handler failed. cleanup */
	free(fd);
	if (procs[tid]->fds == fd)
		procs[tid]->fds = 0;
	else {
		for (f = procs[tid]->fds; f->next != fd; f = f->next) ;
		f->next = 0;
	}

	return r;
}

int bio_close(int tid, int fd)
{
	struct filedes *f, *ff;
	int r;

	/* search filedescriptor info */
	if (!bio_get_fdinfo(tid, fd, &f))
		return -EBADF;

	/* call io handler */
	if ((r = f->ioh->close(f)) < 0)
		return r;

	/* remove filedescriptor */
	free(f);
	if (procs[tid]->fds == f)
		procs[tid]->fds = 0;
	else {
		for (ff = procs[tid]->fds; ff->next != f; ff = ff->next) ;
		ff->next = 0;
	}

	return 0;
}

void bio_clone_fds(int dtid, int stid)
{
	struct filedes *dfd = 0, *sfd;

	for (sfd = procs[stid]->fds; sfd; sfd = sfd->next) {
		if (!dfd) {
			/* first fd cloning. init fd list */
			procs[dtid]->fds = malloc(sizeof(struct filedes));
			dfd = procs[dtid]->fds;
		} else {
			dfd->next = malloc(sizeof(struct filedes));
			dfd = dfd->next;
		}

		if (!sfd->next)
			dfd->next = 0;

		dfd->owner = procs[dtid];
		dfd->mntp = sfd->mntp;
		dfd->ioh = sfd->ioh;
		dfd->fd = sfd->fd;
		dfd->flags = sfd->flags;
		dfd->pos = sfd->pos;

		dfd->ioh->clone_fd(dfd, sfd);
	}
}

void bio_close_all(int tid)
{
	struct filedes *fd = procs[tid]->fds;
	struct filedes *next;

	while (fd) {
		next = fd->next;
		fd->ioh->close(fd);
		free(fd);
		fd = next;
	}
}

int bio_read(int tid, int fd, void *buf, unsigned int count)
{
	struct filedes *f;
	int r;

	/* search filedescriptor info */
	if (!bio_get_fdinfo(tid, fd, &f))
		return -EBADF;

	/* check whether operation is allowed */
	if ((f->flags & O_PERMM) != O_RDWR && (f->flags & O_PERMM) != O_RDONLY)
		return -EBADF;

	/* check whether there is data to read */
	if (!f->ioh->poll(f, POLLIN | POLLRDBAND)) {
		if (f->flags & O_NONBLOCK)
			return -EAGAIN;
		return proc_block_io(sc_ctx, current, WT_READ,
				     POLLIN | POLLRDBAND, f);
	}

	/* call io handler */
	if ((r = f->ioh->read(f, buf, count)) < 0)
		return r;

	/* update position */
	f->pos += r;

	return r;
}

int bio_write(int tid, int fd, const void *buf, unsigned int count)
{
	struct filedes *f;
	int r;

	/* search filedescriptor info */
	if (!bio_get_fdinfo(tid, fd, &f))
		return -EBADF;

	/* check whether operation is allowed */
	if ((f->flags & O_PERMM) != O_RDWR && (f->flags & O_PERMM) != O_WRONLY)
		return -EBADF;

	/* check whether writing is possible now */
	if (!f->ioh->poll(f, POLLOUT | POLLWRBAND)) {
		if (f->flags & O_NONBLOCK)
			return -EAGAIN;
		return proc_block_io(sc_ctx, current, WT_WRITE,
				     POLLOUT | POLLWRBAND, f);
	}

	/* call io handler */
	if ((r = f->ioh->write(f, buf, count)) < 0)
		return r;

	/* update position */
	f->pos += r;

	return r;
}

int bio_get_fdinfo(int tid, int fdi, struct filedes **fds)
{
	struct filedes *f;

	for (f = procs[tid]->fds; f && f->fd != fdi; f = f->next) ;

	if (!f)
		return 0;

	*fds = f;

	return 1;
}
