/* Copyright (c) 2009, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#pragma once

#include "fcntl.h"
#include "stat.h"

struct process;	/* in sched.h */

struct filedes;
struct mountpoint;

/* TODO: be more POSIX conform */

/* Structure of function for handling input/output. Every driver with /dev node
 * and every filesystem implementation has to define such a structure.
 * Functions in there are called by the BIO after the structure arguments are
 * set up.
 * On failure, a negative errno number should be returned in all functions. If
 * not specified elsewise, functions shall return 0 on success.
 * fd is the filedescriptor info struct with which the call was done. The
 * filesystem driver shall only modify its userdata (ud) value. */
struct io_handle {
	/* Called after a mountpoint entry is initalized. This shall setup the
	 * mountpoint user data and shall validate the filesystem type and
	 * state. */
	int (*mount)(struct mountpoint *mntp);

	/* Called after a filedescriptor entry is initalized. It shall look for
	 * the file on the filesystem and initalize the drivers userdata of the
	 * fd to make it possible to do further calls. */
	int (*open)(struct filedes *fd, const char *path);

	/* Shall free the drivers userdata */
	int (*close)(struct filedes *fd);

	/* Copies the userdata from src to dst, so that dst and src point to
	 * exactly the same thing. This is used on fork() when a new process is
	 * created inheriting all filedescriptors from its parent process. */
	void (*clone_fd)(struct filedes *dst, struct filedes *src);

	/* shall read count bytes into buf.
	 * returns the number of bytes read. */
	int (*read)(struct filedes *fd, void *buf, unsigned int count);

	/* shall write count bytes from buf.
	 * returns the number of bytes written. */
	int (*write)(struct filedes *fd, const void *buf, unsigned int count);

	/* shall return something != 0 if this event would be possible */
	int (*poll)(struct filedes *fd, int event);
};

/* Every mountpoint has such a struct. There is one of this linked list on the
 * system containing information about all mountpoints. */
struct mountpoint {
	/* pointer to next mountpoint in this linked list */
	struct mountpoint *next;

	/* filesystem driver functions to handle all I/O requests on this
	 * mountpoint. */
	const struct io_handle *ioh;

	/* path where the mountpoint is mounted to */
	const char *path;

	/* filedes for accessing the source of the mountpoint (the device which
	 * contains the mounted filesystem) */
	struct filedes *src;

	/* userdata for the FD driver to store necessary information about the
	 * mount (superblock, ...) */
	void *ud;
};

/* filedescriptor information struct. Every process has a list with its
 * filedescriptors */
struct filedes {
	/* next filedescriptor of the process in linked list */
	struct filedes *next;

	/* process which opened the file associated with this descriptor */
	struct process *owner;

	/* mountpoint of mount on which this file resides */
	struct mountpoint *mntp;

	/* I/O handling functions of the filedescriptor. These are directly
	 * copied from the mountpoint information on open(). */
	const struct io_handle *ioh;

	/* filedescriptor number which was returned on open() to the process and
	 * is used by the process for referencing this struct */
	int fd;

	/* flags passed on open(). See fcntl.h */
	int flags;

	/* current position in file (relative to start of file) */
	unsigned int pos;

	/* userdata used by the filesystem driver to identify this file (e.g.
	 * inode information, ...) */
	void *ud;
};

extern void bio_init(void);

extern int bio_mount(const char *dev, const char *path, const char *type);
extern int bio_open(int tid, const char *path, int flags, struct filedes **fds);
extern int bio_close(int tid, int fd);
extern void bio_clone_fds(int dtid, int stid);
extern void bio_close_all(int tid);
extern int bio_read(int tid, int fd, void *buf, unsigned int count);
extern int bio_write(int tid, int fd, const void *buf, unsigned int count);

extern int bio_get_fdinfo(int tid, int fdi, struct filedes **fds);

/* default handlers doing nothing */
extern int def_open(struct filedes *fd, const char *path);
extern int def_close(struct filedes *fd);
extern void def_clone_fd(struct filedes *dst, struct filedes *src);
extern int def_poll(struct filedes *fd, int event);
