/* Copyright (c) 2009-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "asm.h"
#include "bio.h"
#include "errno.h"
#include "exec.h"
#include "lib.h"
#include "page.h"
#include "sched.h"

#define EI_NIDENT	16	/* size of e_ident in elf_ehdr */
#define EI_MAG		0	/* magic */
#define EI_CLASS	4	/* class (32 / 64 bit ) */
#define EI_DATA		5	/* endianness */
#define EI_VERSION	6	/* ELF version */

/* right values */
#define E_MAG		("\x7f" "ELF")
#define ELFCLASS32	1
#define ELFDATA2LSB	1
#define EV_CURRENT	1
#define EM_ARM		40
#define ET_EXEC		2
#define ET_DYN		3
#define PT_LOAD		1
#define PF_X		(1 << 0)
#define PF_W		(1 << 1)
#define PF_R		(1 << 2)
#define SHN_UNDEF	0
#define SHT_PROGBITS	1
#define SHT_SYMTAB	2

/* types used in ELF structures */
typedef unsigned long	elf_addr;	/* memory address */
typedef unsigned long	elf_off;	/* file offset */
typedef unsigned short	elf_section;	/* section index */
typedef unsigned short	elf_versym;	/* version symbol information */
typedef	char		elf_byte;
typedef unsigned short	elf_half;
typedef int		elf_sword;
typedef unsigned int	elf_word;
typedef long long	elf_sxword;
typedef unsigned long long elf_xword;

/* start of every ELF file */
struct elf_ehdr {
	elf_byte e_ident[EI_NIDENT];
	elf_half e_type;	/* type of ELF (executable, ...) */
	elf_half e_machine;	/* machine type */
	elf_word e_version;	/* ELF version */
	elf_addr e_entry;	/* entry point */
	elf_off  e_phoff;	/* offset of program header table */
	elf_off  e_shoff;	/* offset of section header table */
	elf_word e_flags;	/* machine flags (unused on ARM) */
	elf_half e_ehsize;	/* ELF header size */
	elf_half e_phentsize;	/* program header entry size */
	elf_half e_phnum;	/* number of program headers */
	elf_half e_shentsize;	/* section header entry size */
	elf_half e_shnum;	/* number of section headers */
	elf_half e_shstrndx;	/* section string table index */
};

/* program header table entry */
struct elf_phdr {
	elf_word p_type;	/* type (loadable, ...) */
	elf_off  p_offset;	/* offset of data */
	elf_addr p_vaddr;	/* address when running (for MMU systems) */
	elf_addr p_paddr;	/* address for loading (for MMU systems) */
	elf_word p_filesz;	/* size in ELF file */
	elf_word p_memsz;	/* size in memory */
	elf_word p_flags;	/* flags (exec, ...) (for MMU systems) */
	elf_word p_align;	/* alignment */
};

/* section header */
struct elf_shdr {
	elf_word sh_name;	/* name (offset in string table) */
	elf_word sh_type;	/* type */
	elf_word sh_flags;	/* flags */
	elf_addr sh_addr;	/* address in RAM */
	elf_off  sh_offset;	/* offset in file */
	elf_word sh_size;	/* length of section */
	elf_word sh_link;	/* additional info (section dependent) */
	elf_word sh_info;	/* additional info (section dependent) */
	elf_word sh_addralign;	/* alignment */
	elf_word sh_entsize;	/* entry size in section (if table) */
};

/* symbol table entry */
struct elf_sym {
	elf_word st_name;	/* name (offset in symbol string table) */
	elf_addr st_value;	/* address of the symbol */
	elf_word st_size;	/* size of the symbol */
	elf_byte st_info;	/* symbol type */
	elf_byte st_other;	/* reserved */
	elf_half st_shndx;	/* section in which this symbol resides */
};

int usrspc_started;

#ifdef CONFIG_SYMTAB
/* return name of symbol at the address */
void elf_symlookup(int tid, unsigned int addr, char *buf)
{
	int i;
	char *name;

	if (!procs[tid]->elfsymtab) {
		strcpy(buf, "(no symbol table)");
		return;
	}

	addr -= 4;	/* to point to branch instruction */
	addr -= (unsigned int) procs[tid]->elfoff;

	for (i = 0; i < procs[tid]->elfsymnum; i++) {
		if (!(procs[tid]->elfsymtab[i].st_value <= addr &&
		      addr <= procs[tid]->elfsymtab[i].st_value +
				procs[tid]->elfsymtab[i].st_size))
			continue;

		name = procs[tid]->elfsymstrtab +
			procs[tid]->elfsymtab[i].st_name;

		if (procs[tid]->elfsymtab[i].st_value == addr) {
			sprintf(buf, "%#x <%s>", addr, name);
			return;
		}

		sprintf(buf, "%#x <%s+%#x>", addr, name,
			addr - (unsigned int) procs[tid]->elfsymtab[i].st_value);
		return;
	}

	strcpy(buf, "(out of ELF)");
}
#endif

int elf_exec(struct context *ctx, const char *path, char **argv, char **envp)
{
	struct elf_ehdr ehdr;
	struct elf_phdr *phdr;
	struct elf_shdr *shdr;
	struct filedes *fd;
	int r, i, l;
	unsigned int j;
	int fdnum;
	void *ph = 0;
	char *strtable;
	void *oldstack, *newstack;
	int argc;
	char **new_argv, **new_envp;
	int cmdlinelen = 0;
	void *picbase = 0;
	int prot;

#ifdef CONFIG_SYMTAB
	procs[current]->elfsymstrtab = 0;
	procs[current]->elfsymtab = 0;
#endif

	/* TODO: be POSIX conform (free pages, init stack, ...) */
	/* TODO: support shared object loading */
	
	/* open ELF file */
	fdnum = bio_open(current, path, O_RDONLY, &fd);
	if (fdnum < 0)
		return fdnum;

	/* read header */
	fd->pos = 0;
	r = fd->ioh->read(fd, &ehdr, sizeof(struct elf_ehdr));
	if (r < 0)
		goto fail;

	/* validate header information */
	if (strncmp(ehdr.e_ident, E_MAG, 4) ||
	    ehdr.e_ident[EI_CLASS] != ELFCLASS32 ||
	    ehdr.e_ident[EI_DATA] != ELFDATA2LSB ||
	    ehdr.e_ident[EI_VERSION] != EV_CURRENT ||
	    (ehdr.e_type != ET_EXEC && ehdr.e_type != ET_DYN) ||
	    ehdr.e_machine != EM_ARM ||
	    ehdr.e_version != EV_CURRENT ||
	    ehdr.e_shstrndx >= ehdr.e_shnum ||
	    ehdr.e_phentsize != sizeof(struct elf_phdr) ||
	    ehdr.e_shentsize != sizeof(struct elf_shdr)) {
		r = -ENOEXEC;
		goto fail;
	}

	/* read & load program headers. as we have no MMU, only one loadable
	 * program header is supported, and no protection flags (exec, read,
	 * write) are supported. */
	phdr = alloca(ehdr.e_phnum * sizeof(struct elf_phdr));
	fd->pos = ehdr.e_phoff;
	r = fd->ioh->read(fd, phdr, ehdr.e_phnum * sizeof(struct elf_phdr));
	if (r < 0)
		goto fail;
	for (i = 0; i < ehdr.e_phnum; i++) {
		/* we only need to look at loadable program headers */
		if (phdr[i].p_type != PT_LOAD)
			continue;

		prot = 0;
		if (phdr[i].p_flags & PF_X)
			prot |= PROT_EXEC;
		if (phdr[i].p_flags & PF_W)
			prot |= PROT_WRITE;
		if (phdr[i].p_flags & PF_R)
			prot |= PROT_READ;

		/* now load */
		ph = page_alloc(current,
				blocks(phdr[i].p_memsz, PAGE_SIZE),
				prot);
		fd->pos = phdr[i].p_offset;
		r = fd->ioh->read(fd, ph,
				  min(phdr[i].p_filesz, phdr[i].p_memsz));
		if (r < 0) {
			page_free(ph, blocks(phdr[i].p_memsz, PAGE_SIZE));
			goto fail;
		}
		if (phdr[i].p_memsz > phdr[i].p_filesz)
			/* zero out .bss (or whatever needs to be zeroed out
			 * here) */
			memset(ph + phdr[i].p_filesz, 0,
			       phdr[i].p_memsz - phdr[i].p_filesz);
		break;
	}
	if (i == ehdr.e_phnum) {
		r = -EINVAL;
		goto fail;
	}

#ifdef CONFIG_SYMTAB
	procs[current]->elfoff = ph;
#endif

	if (ehdr.e_type == ET_DYN) {
		/* dynamic ELF (maybe PIC, as needed for MMU-less systems for
		 * global symbol references). We have to do runtime linking now
		 * to ensure that these references do not cause a data abort
		 * exception. */

		if (ehdr.e_shstrndx == SHN_UNDEF)
			/* no string table... thus runtime linking is
			 * impossible. */
			goto runtime_link_done;

		/* read section table */
		shdr = alloca(ehdr.e_shnum * sizeof(struct elf_shdr));
		fd->pos = ehdr.e_shoff;
		r = fd->ioh->read(fd, shdr,
				  ehdr.e_shnum * sizeof(struct elf_shdr));
		if (r < 0)
			goto fail;

		/* read section name string table */
		strtable = alloca(shdr[ehdr.e_shstrndx].sh_size);
		fd->pos = shdr[ehdr.e_shstrndx].sh_offset;
		r = fd->ioh->read(fd, strtable, shdr[ehdr.e_shstrndx].sh_size);
		if (r < 0) {
			goto fail;
		}

		/* now search till we find proper section */
		for (i = 1; i < ehdr.e_shnum; i++) {
#ifdef CONFIG_SYMTAB
			if (shdr[i].sh_type == SHT_SYMTAB &&
			    !procs[current]->elfsymtab &&
			    !procs[current]->elfsymstrtab &&
			    !strcmp(strtable + shdr[i].sh_name, ".symtab") &&
			    shdr[i].sh_entsize == sizeof(struct elf_sym)) {
				/* read symbol table, but only useful symbols
				 * (starting at sh_info) */
				procs[current]->elfsymtab =
					malloc(shdr[i].sh_size -
					       shdr[i].sh_info *
					       sizeof(struct elf_sym));
				fd->pos = shdr[i].sh_offset +
					shdr[i].sh_info * sizeof(struct elf_sym);
				r = fd->ioh->read(fd, procs[current]->elfsymtab,
						  shdr[i].sh_size -
						  shdr[i].sh_info *
						  sizeof(struct elf_sym));
				if (r < 0)
					goto fail;

				procs[current]->elfsymnum =
					shdr[i].sh_size / sizeof(struct elf_sym) -
					shdr[i].sh_info;

				/* read symbol string table */
				procs[current]->elfsymstrtab =
					malloc(shdr[shdr[i].sh_link].sh_size);
				fd->pos = shdr[shdr[i].sh_link].sh_offset;
				r = fd->ioh->read(fd,
						  procs[current]->elfsymstrtab,
						  shdr[shdr[i].sh_link].sh_size);
				if (r < 0)
					goto fail;
			}
#endif
			/* check */
			if (shdr[i].sh_type != SHT_PROGBITS ||
			    (strcmp(strtable + shdr[i].sh_name, ".got") &&
			     strcmp(strtable + shdr[i].sh_name, ".got.plt") &&
			     strcmp(strtable + shdr[i].sh_name, ".data.rel.local") &&
			     strcmp(strtable + shdr[i].sh_name, ".data.rel.ro.local")))
				continue;

			if (!picbase)
				picbase = ph + shdr[i].sh_addr;

			/* now adjust the pointers */
			for (j = 0; j < shdr[i].sh_size; j += 4)
				*(int *) (ph + shdr[i].sh_addr + j) += (int) ph;
		}
	}
runtime_link_done:

	/* TODO: close all file handles, not just our one */
	bio_close(current, fdnum);

	/* now we set up the ELFs stack, that means, we copy argv and envp
	 * completely to the stack. As our argv and envp arguments are in the
	 * stack of the application, we have to create a new stack */
	oldstack = procs[current]->stack;
	newstack = page_alloc(current, SCHED_PROC_SP, PROT_READ | PROT_WRITE);
	procs[current]->stack = newstack;
#ifdef CONFIG_USRZERO
	memset(newstack, 0, SCHED_PROC_SP * PAGE_SIZE);
#endif
	newstack += SCHED_PROC_SP * PAGE_SIZE;	/* stack end */
	/* copy argv */
	for (l = 0; argv[l]; l++) ;
	argc = l;
	l++;	/* l+1 for terminator */
	new_argv = newstack - 4 * l;
	newstack = new_argv;
	for (i = 0; argv[i]; i++) {
		l = strlen(argv[i]);
		l++;
		cmdlinelen += l;
		newstack -= 4 * blocks(l, 4);
		memcpy(newstack, argv[i], l);
		new_argv[i] = newstack;
	}
	new_argv[i] = 0;
	/* copy envp */
	for (l = 0; envp[l]; l++) ;
	l++;	/* l+1 for terminator */
	new_envp = newstack - 4 * l;
	newstack = new_envp;
	for (i = 0; envp[i]; i++) {
		l = strlen(envp[i]);
		l++;
		newstack -= 4 * blocks(l, 4);
		memcpy(newstack, envp[i], l);
		new_envp[i] = newstack;
	}
	new_envp[i] = 0;

	/* build up cmdline */
	procs[current]->cmdline = realloc(procs[current]->cmdline, cmdlinelen);
	memcpy(procs[current]->cmdline, argv[0], strlen(argv[0]));
	l = strlen(argv[0]);
	for (i = 1; argv[i]; i++) {
		procs[current]->cmdline[l++] = ' ';
		memcpy(procs[current]->cmdline + l, argv[i], strlen(argv[i]));
		l += strlen(argv[i]);
	}
	procs[current]->cmdline[l] = 0;

	/* TODO: free old pages (not only stack, etc.) */
	page_free(oldstack, SCHED_PROC_SP);

	/* reset signal handlers to initial values (former function pointers are
	 * invalid now) */
	sig_reset_handlers(current);

	usrspc_started = 1;
	procs[current]->img_replaced = 1;

	/* jump into the programm
	 * TODO: support binaries with THUMB instructions */
#ifdef CONFIG_USRZERO
	memset(ctx, 0, sizeof(struct context));
#endif
	ctx->psr = PSR_M_usr | PSR_F;
	ctx->r0  = argc;
	ctx->r1  = (int) new_argv;
	ctx->r2  = (int) new_envp;
	ctx->sl  = (int) picbase;
	ctx->sp  = (int) newstack;
	ctx->pc  = (int) (ph + ehdr.e_entry);
	return DONT_RETURN;

fail:
#ifdef CONFIG_SYMTAB
	if (procs[current]->elfsymtab)
		free(procs[current]->elfsymtab);
	if (procs[current]->elfsymstrtab)
		free(procs[current]->elfsymstrtab);
#endif
	bio_close(current, fdnum);
	return r;
}
