/*-
 * Copyright (c) 2010 Alexander Graf.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "bio.h"
#include "errno.h"
#include "lib.h"

#include "ext2fs.h"

/* on-memory per-mount filesystem information */
struct fsinf {
	struct ext2_super_block sb;
	int groups;
	struct ext2_group_desc *gd;
};

/* on-memory per-filedescriptor information */
struct fdinf {
	struct ext2_inode ino;
};

/* maximum steps in a blockpath */
#define EXT2_BLOCKPATH	4

static int ext2fs_mount(struct mountpoint *mp)
{
	int r;
	struct fsinf *fsinf = 0;

	mp->ud = malloc(sizeof(struct fsinf));
	fsinf = mp->ud;
	fsinf->gd = 0;

	/* read superblock.  First block is reserved for bootloader, second
	 * block is superblock. */
	mp->src->pos = 1024;
	r = mp->src->ioh->read(mp->src, &fsinf->sb,
			       sizeof(struct ext2_super_block));
	if (r < 0)
		goto err;

	if (fsinf->sb.s_magic != EXT2_SB_MAGIC) {
		printf("ext2fs: not an ext2 filesystem.\n");
		r = -EINVAL;
		goto err;
	}

	/* check state, but just warn if it is not valid. */
	if (fsinf->sb.s_state != EXT2_VALID_FS)
		printf("ext2fs: filesystem not cleanly unmounted.\n");

	fsinf->groups = blocks(fsinf->sb.s_blocks_count,
			       fsinf->sb.s_blocks_per_group);
	fsinf->gd = malloc(fsinf->groups * sizeof(struct ext2_group_desc));

	/* read group descriptors.  They are in the block behind the superblock
	 * (third block) */
	mp->src->pos = 2048;
	r = mp->src->ioh->read(mp->src, fsinf->gd,
			       fsinf->groups * sizeof(struct ext2_group_desc));
	if (r < 0)
		goto err;

	return 0;

err:
	if (fsinf) {
		if (fsinf->gd)
			free(fsinf->gd);
		free(fsinf);
	}
	return r;
}

static int ext2fs_get_inode(struct mountpoint *mp, int nr, struct ext2_inode *ino)
{
	struct fsinf *fsinf = mp->ud;
	int gnr = (nr-1) / fsinf->sb.s_inodes_per_group;
	int inr = (nr-1) % fsinf->sb.s_inodes_per_group;
	int iaddr;

	iaddr  = fsinf->gd[gnr].bg_inode_table * EXT2_BLOCK_SIZE(fsinf->sb);
	iaddr += inr * EXT2_INODE_SIZE(fsinf->sb);

	mp->src->pos = iaddr;
	return mp->src->ioh->read(mp->src, ino, sizeof(struct ext2_inode));
}

static int ext2fs_get_blockpath(struct mountpoint *mp, unsigned int block,
				int *path)
{
	struct fsinf *fsinf = mp->ud;

	if (block < EXT2_NDIR_BLOCKS) {
		/* direct block */
		*path = block;
		return 1;
	} else if ((block -= EXT2_NDIR_BLOCKS) <
		   EXT2_ADDR_PER_BLOCK(fsinf->sb)) {
		/* indirect block */
		*path++ = EXT2_IND_BLOCK;
		*path   = block;
		return 2;
	} else if ((block -= EXT2_ADDR_PER_BLOCK(fsinf->sb)) <
		   square(EXT2_ADDR_PER_BLOCK(fsinf->sb))) {
		/* double indirect block */
		*path++ = EXT2_DIND_BLOCK;
		*path++ = block / EXT2_ADDR_PER_BLOCK(fsinf->sb);
		*path   = block % EXT2_ADDR_PER_BLOCK(fsinf->sb);
		return 3;
	} else {
		/* triple indirect block */
		block -= square(EXT2_ADDR_PER_BLOCK(fsinf->sb));
		*path++ = EXT2_TIND_BLOCK;
		*path++ = block / square(EXT2_ADDR_PER_BLOCK(fsinf->sb));
		block %= square(EXT2_ADDR_PER_BLOCK(fsinf->sb));
		*path++ = block / EXT2_ADDR_PER_BLOCK(fsinf->sb);
		*path   = block % EXT2_ADDR_PER_BLOCK(fsinf->sb);
		return 4;
	}
}

static int ext2fs_get_block(struct mountpoint *mp, struct ext2_inode *ino,
			    int block)
{
	struct fsinf *fsinf = mp->ud;
	int addr;
	int blockpath[EXT2_BLOCKPATH];
	int bpl;
	int i;
	int r;

	bpl = ext2fs_get_blockpath(mp, block, blockpath);

	addr = ino->i_block[blockpath[0]];

	for (i = 1; i < bpl; i++) {
		mp->src->pos  = addr * EXT2_BLOCK_SIZE(fsinf->sb);
		mp->src->pos += blockpath[i] * EXT2_ADDR_SIZE;
		r = mp->src->ioh->read(mp->src, &addr, EXT2_ADDR_SIZE);
		if (r < 0)
			return r;
	}

	return addr;
}

static int ext2fs_read_data(struct mountpoint *mp, struct ext2_inode *ino,
			    unsigned int addr, void *buf, unsigned int count)
{
	struct fsinf *fsinf = mp->ud;
	int block;
	int start;
	unsigned int size;
	int rv = 0;
	int r;

	if (addr + count > ino->i_size)
		count = ino->i_size - addr;
	if (count <= 0)
		return 0;

	while (count) {
		block = ext2fs_get_block(mp, ino,
					 addr / EXT2_BLOCK_SIZE(fsinf->sb));
		if (block < 0)
			return block;

		start = (block * EXT2_BLOCK_SIZE(fsinf->sb)) +
			(addr % EXT2_BLOCK_SIZE(fsinf->sb));

		size = EXT2_BLOCK_SIZE(fsinf->sb) -
			(addr % EXT2_BLOCK_SIZE(fsinf->sb));
		size = min(size, count);

		mp->src->pos = start;
		r = mp->src->ioh->read(mp->src, buf + rv, size);
		if (r < 0)
			return r;

		addr  += size;
		rv    += size;
		count -= size;
	}

	return rv;
}

#define CD(de, s, l)	(de.name_len == l && strncmp(de.name, s, l) == 0)

static int ext2fs_search(struct mountpoint *mp, const char *path)
{
	int ino;	/* current inode */
	struct ext2_inode in;
	struct ext2_dir_entry de;
	const char *p;	/* path - already processed parts skipped */
	int ctns;	/* chars to next slash */
	int last;	/* 1 if looking for last part of pathname */
	int addr;	/* address in directory */
	void *tmp;
	int r;

	ino = EXT2_ROOT_INO;
	p = path;
	last = 0;

search:
	tmp = strchr(p, '/');
	if (!tmp) {
		last = 1;
		ctns = strlen(p);
	} else {
		ctns = (unsigned int) tmp - (unsigned int) p;
	}
	if (!ctns)
		return ino;

	r = ext2fs_get_inode(mp, ino, &in);
	if (r < 0)
		return r;

	/* TODO: not fast */

	addr = 0;
	while ((r = ext2fs_read_data(mp, &in, addr, &de,
				     sizeof(struct ext2_dir_entry))) > 0) {
		addr += de.rec_len;

		if (!CD(de, p, ctns))
			continue;
		if (last)
			return de.inode;
		if (de.file_type != EXT2_FT_DIR)
			return -ENOTDIR;

		ino = de.inode;
		p += ctns + 1;
		goto search;
	}

	if (r < 0)
		return r;
	return -ENOENT;
}

static int ext2fs_open(struct filedes *fd, const char *path)
{
	struct fdinf *fdinf;
	int r;

	r = ext2fs_search(fd->mntp, path);
	if (r < 0)
		return r;

	fd->ud = malloc(sizeof(struct fdinf));
	fdinf = fd->ud;

	r = ext2fs_get_inode(fd->mntp, r, &fdinf->ino);
	if (r < 0)
		free(fdinf);

	return r;
}

static int ext2fs_close(struct filedes *fd)
{
	free(fd->ud);
	return 0;
}

static void ext2fs_clone_fd(struct filedes *dst, struct filedes *src)
{
	dst->ud = malloc(sizeof(struct fdinf));
	memcpy(dst->ud, src->ud, sizeof(struct fdinf));
}

static int ext2fs_read(struct filedes *fd, void *buf, unsigned int count)
{
	struct fdinf *fdinf = fd->ud;

	return ext2fs_read_data(fd->mntp, &fdinf->ino, fd->pos, buf, count);
}

const struct io_handle ext2fs_ioh = {
	.mount		= ext2fs_mount,
	.open		= ext2fs_open,
	.close		= ext2fs_close,
	.clone_fd	= ext2fs_clone_fd,
	.read		= ext2fs_read,
	.poll		= def_poll
};
