/* Copyright (c) 2009-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "asm.h"
#include "compiler.h"
#include "exec.h"
#include "lib.h"
#include "sched.h"
#include "page.h"
#include "pf/cpu.h"

#define SYMBUFSIZ	64

#ifdef CONFIG_STACKTRACE
static void stacktrace(int *fp, int tid)
{
	int i = 0;
	int lr, nextfp;
#ifdef CONFIG_SYMTAB
	char symb[SYMBUFSIZ];
#endif

	puts("\nCall stack:");

	while (is_in_ram(fp)) {
		fp--;

		nextfp = fp[0];
		lr = fp[1];

#ifdef CONFIG_SYMTAB
		if (tid) {
			elf_symlookup(tid, lr, symb);
			printf("%3i: fp = %#010x, lr = %#010x: %s\n",
			       ++i, (int) fp, lr, symb);
		} else
#endif
		printf("%3i: fp = %#010x, lr = %#010x\n", ++i, (int) fp, lr);

		if ((int) fp == nextfp - 4)
			/* avoid getting into endless loop by faked stacks */
			break;

		fp = (int *) nextfp;
	}

	printf("%3i: fp = %#010x (illegal)\n", ++i, (int) fp);
}
#endif

/* declaration to shut up compiler */
extern void _kpanic(int *fp, const char *s, ...);

#ifdef CONFIG_STACKTRACE
__noreturn __used __printf(2, 3) _kpanic(int *fp, const char *s, ...)
#else
void kpanic(const char *s, ...)
#endif
{
	va_list ap;

	led_indicate_error();

	printf("\nKernel panic: ");

	va_start(ap, s);
	vprintf(s, ap);
	va_end(ap);

#ifdef CONFIG_STACKTRACE
	puts("");
	stacktrace(fp, 0);
#endif

	puts("\nSystem halted.");

	/* TODO: use very deep sleep mode */
	idle();
}

void exception(struct context *ctx, int tid, int sig)
{
	int i;
	unsigned int *g = (unsigned int *) ctx;
#ifdef CONFIG_SYMTAB
	char symb[SYMBUFSIZ];
#endif

	/* TODO: dump signal name */
	printf("\nfatal signal %i sent to process with tid %i, pid %i\n",
	       sig, tid, procs[tid]->pid);

	puts("\nRegister dump:");

	printf(" psr = %#010x\n", g[0]);
	g++;
	for (i = 0; i < 4; i++)
		printf(" r%02i = %#010x    r%02i = %#010x   "
		       " r%02i = %#010x    r%02i = %#010x\n",
		       i*4+0, g[i*4+0], i*4+1, g[i*4+1],
		       i*4+2, g[i*4+2], i*4+3, g[i*4+3]);

#ifdef CONFIG_SYMTAB
	if (tid) {
		puts("\nSymbols:");
		elf_symlookup(tid, ctx->lr, symb);
		printf(" lr: %s\n", symb);
		elf_symlookup(tid, ctx->pc, symb);
		printf(" pc: %s\n", symb);
	}
#endif

#ifdef CONFIG_STACKTRACE
	stacktrace((int *) ctx->fp, tid);
#endif

	puts("\nexiting process.");

	/* TODO: be able to dump core */
}

/* shut up compiler, as there is no previous definition of them because it is
 * not needed, since they are called by assembler code (isr.S) */
extern void exc_undefined(struct context *ctx);
extern void exc_prefetch(struct context *ctx);
extern void exc_data(struct context *ctx);

void exc_undefined(struct context *ctx)
{
	sig_emit(ctx, current, SIGILL);
}

void exc_prefetch(struct context *ctx)
{
	sig_emit(ctx, current, SIGSEGV);
}

void exc_data(struct context *ctx)
{
	sig_emit(ctx, current, SIGSEGV);
}
