/* Copyright (c) 2009, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "compiler.h"
#include "dev.h"
#include "bio.h"
#include "lib.h"

/* format of partition entry */
struct part {
	unsigned char bootind;		/* bit 7 indicates whether bootable */
	unsigned char start_chs[3];	/* start cylinder/head/sector */
	unsigned char sysind;		/* type of partition */
	unsigned char end_chs[3];	/* end cylinder/head/sector */
	unsigned int start;		/* number of first sector */
	unsigned int size;		/* partition size in sectors */
} __packed;

/* format of partition table */
struct pt {
	struct part part[4];
	unsigned short magic;
} __packed;

/* address of partition table */
#define PT_ADDR		0x1be

/* magic */
#define PT_MAGIC	0xaa55

struct ibmpart_ud {
	const struct io_handle *ioh;
	unsigned int start;
};

static void ibmpart_clone_fd(struct filedes *dst, struct filedes *src)
{
	/* no copying needed, as there is one userdata pointer per ibmpart
	 * device. Every filedes on one device get the same pointer. */
	dst->ud = src->ud;
}

static int ibmpart_read(struct filedes *fd, void *buf, unsigned int count)
{
	int rv;
	struct ibmpart_ud *ud = (struct ibmpart_ud *) fd->ud;

	fd->pos += ud->start;
	rv = ud->ioh->read(fd, buf, count);
	fd->pos -= ud->start;

	return rv;
}

static int ibmpart_write(struct filedes *fd, const void *buf,
			 unsigned int count)
{
	int rv;
	struct ibmpart_ud *ud = (struct ibmpart_ud *) fd->ud;

	fd->pos += ud->start;
	rv = ud->ioh->write(fd, buf, count);
	fd->pos -= ud->start;

	return rv;
}

static const struct io_handle ibmpart_ioh = {
	.open		= def_open,
	.close		= def_close,
	.clone_fd	= ibmpart_clone_fd,
	.read		= ibmpart_read,
	.write		= ibmpart_write,
	.poll		= def_poll
};

void ibmpart_scan(const char *prefix, const struct io_handle *ioh)
{
	struct pt pt;
	int i;
	struct filedes fd;
	struct ibmpart_ud *ud;
	char *name;

	fd.ioh = ioh;
	fd.pos = PT_ADDR;
	i = fd.ioh->read(&fd, &pt, sizeof(pt));

	if (i < 0) {
		printf("%s: cannot read partition table, errno %i\n",
		       prefix, i);
		return;
	}

	if (i != sizeof(pt) || pt.magic != PT_MAGIC)
		return;

	/* TODO: support logical stuff */

	for (i = 0; i < 4; i++) {
		if (pt.part[i].sysind) {
			ud = malloc(sizeof(struct ibmpart_ud));
			ud->ioh = ioh;
			ud->start = pt.part[i].start << 9;
			name = malloc(strlen(prefix) + 3);
			sprintf(name, "%s.%i", prefix, i);
			devfs_add(name, &ibmpart_ioh, ud);
			printf("%s: partition size = %i MB, type = %#x\n",
			       name, pt.part[i].size >> 11,
			       (int) pt.part[i].sysind);
		}
	}
}
