/* Copyright (c) 2008-2009, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "bio.h"
#include "errno.h"
#include "lib.h"

/* The minix filesystem is build up of:
 *
 * 1 kb nothing (reserverd for bootloader etc.)
 * 1 kb superblock
 * (imap_blocks * BLOCKSIZE) bytes inode map
 * (zmap_blocks * BLOCKSIZE) bytes zone map
 * ninodes * sizeof(struct inode) bytes inode table
 * up to (ZONESIZE - 1) bytes nothing (to fill the current zone)
 * (nzones - firstdatazone) * ZONESIZE bytes data zones */

/* Some words about the inodes, zones and blocks:
 *
 * - The 0. inode is not used. It has its entry (set to 1) in the inode bitmap,
 *   but it has NO entry in the inode table.
 * - The root directory has always inode #1
 * - inode::zone uses the first 7 entries to store the numbers of the first 7
 *   zones in which the file is (direct zone). The 8. entry (inode::zone[7])
 *   stores the number of a zone containing this array for the next zones of the
 *   file (indirect zone). The 9. entry (inode::zone[8]) stores the zone number
 *   of a zone containing an array of zones like zone[7] (double indirect zone)
 *
 * - A block has a size of BLOCKSIZE bytes.
 *
 * - A zone has a size of ZONESIZE bytes.
 * - The WHOLE filesystem is seperated into zones. Not only the data area.
 *   superblock::firstdatazone points to the first zone usable for data (The
 *   zone after the last zone used for the inode table)
 * - The first unusable zones (the first superblock::firstdatazone zones) have
 *   NO entry in the zone bitmap.
 *
 * - Both zone and inode maps are bitmaps indicating whether the zone/inode is
 *   used or not. 1 = used, 0 = unused.
 * - To check, whether inode 'i' is used, do:
 *   (map[i / 8] & 1 << (i % 8)).
 * - To check, whether zone 'z' is used, do:
 *   z -= superblock::firstdatazone;
 *   (map[z / 8] & 1 << (z % 8)), because the unusable zones have no entry in
 *   the zone bitmap. */

/* Directories:
 *
 * In the data area of directory inodes is an array of dirents of the
 * directories/files stored in the directory. */

/* magic for minix V1 filesystem with 30 character long filenames */
#define V1_MAGIC	0x138f

/* valid values for superblock::state */
#define STATE_CLEAN	1
#define STATE_DIRTY	2

#define BLOCKSIZE	1024

#define ZONESIZE(sb)	(BLOCKSIZE << sb.log_zone_size)

/* maximal number of links needed to find the file. This is
 * implementation-defined and we use it to detect endless loops on filesystems
 *
 * e.g. if A is a link to B, B is a link to C and C is a link to A, we would
 * recurse (really RECURSE, not iterate) in an endless loop when trying to
 * access A, B or C.
 *
 * After MAX_LINKS are reached, we return ELOOP. */
#define MAX_LINKS	10

/* on-disk superblock information */
struct superblock {
	unsigned short ninodes;		/* usable inodes		*/
	unsigned short nzones;		/* total number of zones	*/
	unsigned short imap_blocks;	/* size of inode map [blocks]	*/
	unsigned short zmap_blocks;	/* size of zone map  [blocks]	*/
	unsigned short firstdatazone;	/* number of first data zone	*/
	unsigned short log_zone_size;	/* log2 of blocks/zone		*/
	unsigned long  max_size;	/* maximum file size		*/
	unsigned short magic;		/* magic number			*/
	unsigned short state;		/* file system state		*/
};

/* on-disk inode information */
struct inode {
	unsigned short mode;		/* file type, protection, etc.	*/
	unsigned short uid;		/* user id of the file's owner	*/
	unsigned long  size;		/* current file size in bytes	*/
	unsigned long  mtime;		/* modification time		*/
	unsigned char  gid;		/* group id of the file's owner	*/
	unsigned char  nlinks;		/* number of links to this file	*/
	unsigned short zone[9];		/* zone numbers			*/
};

/* on-disk directory entry information (link to file) */
struct dent {
	unsigned short ino;		/* inode number of the dirent	*/
	char name[30];			/* filename			*/
};

/* on-memory per-mount filesystem information */
struct fsinf {
	struct superblock sb;
	int readonly;
};

/* on-memory per-filedes information */
struct fdinf {
	struct inode ino;
};

static int minix_mount(struct mountpoint *mntp)
{
	struct fsinf *fsinf;

	mntp->ud = malloc(sizeof(struct fsinf));
	fsinf = mntp->ud;

	fsinf->readonly = 0;

	/* read superblock, starting at second block */
	mntp->src->pos = BLOCKSIZE;
	mntp->src->ioh->read(mntp->src, &fsinf->sb, sizeof(struct superblock));

	if (fsinf->sb.magic != V1_MAGIC) {
		printf("minix: wrong/unsupported filesystem magic: 0x%x\n",
		       fsinf->sb.magic);
		return -EINVAL;
	}

	if (fsinf->sb.state != STATE_CLEAN && fsinf->sb.state != STATE_DIRTY) {
		printf("minix: unknown filesystem state: %i\n",
		       fsinf->sb.state);
		return -EINVAL;
	}

	if (fsinf->sb.state == STATE_DIRTY) {
		printf("minix: filesystem is marked as dirty\n");
		fsinf->readonly = 1;
	}

	return 0;
}

static int minix_get_inode_pos(struct mountpoint *mntp, int nr)
{
	struct fsinf *fsinf = mntp->ud;
	int pos = 0;

	/* first two blocks are unused and superblock */
	pos += BLOCKSIZE * 2;

	/* inode and zone bitmaps */
	pos += BLOCKSIZE * fsinf->sb.imap_blocks;
	pos += BLOCKSIZE * fsinf->sb.zmap_blocks;

	/* and now the real inode (inode 0 has no entry in the table) */
	pos += (nr - 1) * sizeof(struct inode);

	return pos;
}

/* TODO: do not ignore return value of this function */
static int minix_get_inode(struct mountpoint *mntp, int nr, struct inode *ino)
{
	mntp->src->pos = minix_get_inode_pos(mntp, nr);
	return mntp->src->ioh->read(mntp->src, ino, sizeof(struct inode));
}

/* stores the path needed to go to get the physical zone number of the given
 * logical zone number */
static void minix_blockpath(struct mountpoint *mntp, int zone, int *path)
{
	int i = 0;
	struct fsinf *fsinf = mntp->ud;

	if (zone < 7) {
		/* direct zone */
		path[i++] = zone;
	} else if ((zone -= 7) < ZONESIZE(fsinf->sb) / 2) {
		/* indirect zone */
		path[i++] = 7;
		path[i++] = zone;
	} else {
		/* double indirect zone */
		zone -= ZONESIZE(fsinf->sb) / 2;
		path[i++] = 8;
		path[i++] = zone / (ZONESIZE(fsinf->sb) / 2);
		path[i++] = zone % (ZONESIZE(fsinf->sb) / 2);
	}
}

/* returns physical zone number of the logical zone number. */
static int minix_get_zone(struct mountpoint *mntp, int logical,
			  const struct inode *ino)
{
	int addr;
	int path[3];
	struct fsinf *fsinf = mntp->ud;
	unsigned short rv;

	minix_blockpath(mntp, logical, path);

	if (path[0] < 7)
		return ino->zone[path[0]];

	addr  = ino->zone[path[0]] * ZONESIZE(fsinf->sb);
	addr += path[1] * 2;
	mntp->src->pos = addr;
	mntp->src->ioh->read(mntp->src, &rv, 2);

	if (path[0] == 7)
		return rv;

	addr  = rv * ZONESIZE(fsinf->sb);
	addr += path[2] * 2;
	mntp->src->pos = addr;
	mntp->src->ioh->read(mntp->src, &rv, 2);

	return rv;
}

static int minix_read_data(struct mountpoint *mntp, const struct inode *ino,
			   unsigned int addr, void *buf, int count)
{
	int zone, blocksize;
	int rb = 0;
	int blockaddr;
	int r;
	struct fsinf *fsinf = mntp->ud;

	if (addr + count > ino->size)
		/* forbid reading after the file */
		count = ino->size - addr;
	if (count <= 0)
		/* nothing to read */
		return 0;

	while (count) {
		/* get zone number of the data we are going to read */
		zone = minix_get_zone(mntp, addr / ZONESIZE(fsinf->sb), ino);

		/* the size of the next block we read is the size of the area
		 * between our current start and the end of the zone, or our
		 * count if it is lower. */
		blocksize = ZONESIZE(fsinf->sb) - (addr % ZONESIZE(fsinf->sb));
		blocksize = min(blocksize, count);

		/* calculate the addres of the zone + start of requestd area */
		blockaddr = (zone * ZONESIZE(fsinf->sb)) +
			(addr % ZONESIZE(fsinf->sb));

		/* read blocksize bytes */
		mntp->src->pos = blockaddr;
		r = mntp->src->ioh->read(mntp->src, buf + rb, blocksize);
		if (r < 0)
			return r;

		addr  += blocksize;	/* start position of the block */
		rb    += blocksize;	/* number of already read bytes */
		count -= blocksize;	/* bytes we still have to read */
	}

	return rb;
}

static int minix_search(struct mountpoint *mntp, const char *file,
			const struct inode *ino, int links)
{
	struct inode ni;	/* next inode */
	struct dent dents[16];	/* 16 * sizeof(struct dent) = 512 = 1 block. */
	int i = 0, rv, c;
	unsigned int ctns;	/* chars to next slash */
	int sf = 0;		/* 1 if we search last file in path */
	struct dent *dent;
	char *cr;

	if (!links--)
		return -ELOOP;

	/* skip first slash */
	if (*file == '/')
		file++;

	/* search next slash */
	cr = strchr(file, '/');
	if (!cr) { /* no slash: we search the last file in the path */
		ctns = strlen(file);
		sf = 1;
	} else if (cr == file)	/* '//' in filename. */
		return -EINVAL;
	else
		ctns = cr - file;

	do {
		rv = minix_read_data(mntp, ino, i, dents, sizeof(dents));
		if (rv < 0)
			return rv;

		i += rv;

		for (c = 0, dent = &dents[0];
		     c < rv;
		     c += sizeof(struct dent), dent++) {
			if (!strncmp(file, dent->name, ctns) &&
			    !dent->name[ctns]) {
				/* found matching file */
				if (sf)
					return dent->ino;
				minix_get_inode(mntp, dent->ino, &ni);
				if (!S_ISDIR(ni.mode))
					return -ENOTDIR;
				else if (!file[ctns+1])
					return dent->ino;
				else
					return minix_search(mntp, file + ctns,
							    &ni, links);
			}
		}
	} while (rv == sizeof(dents));

	/* the while loop should never end this way */
	return -ENOENT;
}

static int minix_open(struct filedes *fd, const char *path)
{
	struct inode ino;
	struct fdinf *fdinf;
	int i;

	fd->ud = malloc(sizeof(struct fdinf));
	fdinf = fd->ud;

	/* this must be here since minix_search does not work with '/' as
	 * parameter */
	if (path[0] == 0) {
		minix_get_inode(fd->mntp, 1, &fdinf->ino);
		return 0;
	}

	/* searching at files always starts at the root inode */
	minix_get_inode(fd->mntp, 1, &ino);

	i = minix_search(fd->mntp, path, &ino, MAX_LINKS);
	if (i < 0) {
		free(fdinf);
		return i;
	}

	minix_get_inode(fd->mntp, i, &fdinf->ino);
	return 0;
}

static int minix_close(struct filedes *fd)
{
	free(fd->ud);
	return 0;
}

static void minix_clone_fd(struct filedes *dst, struct filedes *src)
{
	dst->ud = malloc(sizeof(struct fdinf));
	memcpy(dst->ud, src->ud, sizeof(struct fdinf));
}

static int minix_read(struct filedes *fd, void *buf, unsigned int count)
{
	struct fdinf *fdinf = fd->ud;

	return minix_read_data(fd->mntp, &fdinf->ino, fd->pos, buf, count);
}

const struct io_handle minix_ioh = {
	.mount		= minix_mount,
	.open		= minix_open,
	.close		= minix_close,
	.clone_fd	= minix_clone_fd,
	.read		= minix_read,
	.poll		= def_poll
	/* TODO: .write */
};
