/* Copyright (c) 2008-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "bio.h"
#include "dev.h"
#include "devices.h"
#include "errno.h"
#include "lib.h"
#include "pf/cpu.h"
#include "pf/mmc.h"

/* TODO: have generic, more flexible SPI implementation */

#define HIGH_SPEED_MODE	0	/* full speed */
#define LOW_SPEED_MODE	1	/* speed lowered with waiting loops */

/* maximum bits per second in low speed mode */
#define LOW_SPEED_BPS	(150 << 10)

/* wait cycles for low speed mode
 * is divided by 4 because there are 2 waiting loops, and at least 2
 * instructions are needed for such a loop */
#define LOW_SPEED_WCYC	(CPU_CLOCK / LOW_SPEED_BPS / 4)

static void __optimize sspi_write(unsigned char byte, int mode)
{
	volatile int t;
	int i;

	for (i = 7; i >= 0; i--) {
		if (mode == LOW_SPEED_MODE)
			for (t = LOW_SPEED_WCYC; t; t--) ;

		if (byte & (1 << i))
			IOSET = DI;
		else
			IOCLR = DI;

		IOCLR = CLK;

		if (mode == LOW_SPEED_MODE)
			for (t = LOW_SPEED_WCYC; t; t--) ;

		IOSET = CLK;
	}

	IOSET = DI;
}

static unsigned char __optimize sspi_read(int mode)
{
	volatile int t;
	int i;
	unsigned char rv = 0;

	/* since we always keep DI on high when we do nothing with it, this
	 * function also sends the 0xff dummy byte. */

	for (i = 7; i >= 0; i--) {
		if (mode == LOW_SPEED_MODE)
			for (t = LOW_SPEED_WCYC; t; t--) ;

		IOCLR = CLK;

		if (mode == LOW_SPEED_MODE)
			for (t = LOW_SPEED_WCYC; t; t--) ;

		if (IOPIN & DO)
			rv |= (1 << i);
		IOSET = CLK;
	}

	return rv;
}

/* MMC commands */
#define GO_IDLE_STATE		0	/* resets the MMC */
#define SEND_OP_COND		1	/* activates the MMC' initalization process */
#define READ_SINGLE_BLOCK	17	/* reads a single block */
#define WRITE_BLOCK		24	/* writes a block */

#define BLOCKSIZE	512

/* number of times, something should be retried. */
#define TIMEOUT	200

/* TODO: have generic blocklayer with buffer/cache support */

/* this function sends a command and leaves CS active after it. It returns the
 * answer or -EIO when the card times out */
static int mmc_command(int cmd, int argument, int mode)
{
	unsigned char instruction[6];
	unsigned char rv;
	int i;
	int timeout;

	/* the instructions have the following layout:
	 * 01, 6 bit command, 32 bit argument, 7 bit CRC, 1
	 *
	 * In SPI mode, the CRC needs not to be valid. Nevertheless, we have to
	 * send it when we sending the GO_IDLE_STATE command, since we are in
	 * MMC-Interface mode, when we do so (It's the command which switches to
	 * SPI). */
	instruction[0] = 0x40 | cmd;
	instruction[1] = (argument & 0xff000000) >> 24;
	instruction[2] = (argument & 0x00ff0000) >> 16;
	instruction[3] = (argument & 0x0000ff00) >> 8;
	instruction[4] = (argument & 0x000000ff);
	instruction[5] = (cmd == GO_IDLE_STATE) ? 0x95 : 0xff;

	/* send a dummy byte */
	IOSET = CS;
	sspi_write(0xff, mode);

	/* send the instruction */
	IOCLR = CS;
	for (i = 0; i < 6; i++)
		sspi_write(instruction[i], mode);

	/* wait for a valid answer */
	for (rv = 0xff, timeout = TIMEOUT;
	     rv == 0xff && timeout;
	     rv = sspi_read(mode), timeout--) ;
	if (!timeout)
		return -EIO;

	return rv;
}

/* reads a block */
static int mmc_read_block(unsigned int start, unsigned int count, void *buf)
{
	unsigned int i;
	int timeout;
	int r;
	unsigned int block = start / BLOCKSIZE;
	unsigned int off   = start % BLOCKSIZE;
	char c;

	/* TODO: do not use READ_SINGLE_BLOCK.  This could avoid the need of
	 * mmc_read() function splitting into block chunks.  */

	if ((r = mmc_command(READ_SINGLE_BLOCK, block * BLOCKSIZE,
			     HIGH_SPEED_MODE)) < 0)
		return r;

	if (r != 0)
		/* card sent error code
		 * TODO: evaluate it and printf it */
		return -EIO;

	/* wait for start byte */
	for (timeout = TIMEOUT;
	     sspi_read(HIGH_SPEED_MODE) != 0xfe && timeout;
	     timeout--) ;
	if (!timeout)
		return -EIO;

	for (i = 0; i < BLOCKSIZE; i++) {
		c = sspi_read(HIGH_SPEED_MODE);
		if (off <= i && i < off + count)
			*((char *) buf++) = c;
	}

	/* ignore 16 bit CRC sum */
	sspi_read(HIGH_SPEED_MODE);
	sspi_read(HIGH_SPEED_MODE);

	IOSET = CS;

	return count;
}

static int mmc_read(struct filedes *fd, void *buf, unsigned int count)
{
	int r;
	unsigned int cc;
	unsigned int j = 0, pos = fd->pos;

	while (j < count) {
		/* Three cases must be considered for setting cc:
		 * 1. current addr might be behind the beginning of a block
		 * 2. current addr might be at the beginning of a block
		 * 3. current addr might be in the last requested block */
		cc = BLOCKSIZE - pos % BLOCKSIZE;	/* 1 */
		if (!cc)				/* 2 */
			cc = BLOCKSIZE;
		if (pos + cc > fd->pos + count)		/* 3 */
			cc = fd->pos + count - pos;

		if ((r = mmc_read_block(pos, cc, buf + j)) < 0)
			return r;

		pos += r;
		j   += r;
	}

	return j;
}

/* writes block */
static int mmc_write_block(unsigned int block, const void *buf)
{
	int i, timeout;
	int r;

	if ((r = mmc_command(WRITE_BLOCK, block * BLOCKSIZE,
			     HIGH_SPEED_MODE)) < 0)
		return r;

	if (r != 0)
		/* card sent error code
		 * TODO: evaluate it and printf it */
		return -EIO;

	/* send dummy bytes
	 * TODO: read from CSD or datasheet, how many dummy bytes we actually
	 * need to send */
	for (i = 0; i < 15; i++)
		sspi_write(0xff, HIGH_SPEED_MODE);

	/* send start byte */
	sspi_write(0xfe, HIGH_SPEED_MODE);

	/* send data */
	for (i = 0; i < BLOCKSIZE; i++)
		sspi_write(*((const char *) buf++), HIGH_SPEED_MODE);

	/* send wrong CRC sum */
	sspi_write(0xff, HIGH_SPEED_MODE);
	sspi_write(0xff, HIGH_SPEED_MODE);

	/* wait till the card is done (no busy tokens anymore) */
	for (timeout = TIMEOUT;
	     sspi_read(HIGH_SPEED_MODE) != 0xff && timeout;
	     timeout--) ;
	if (!timeout)
		return -EIO;

	IOSET = CS;
	
	return 0;
}

/* writes a part of the block, reads remainder-part of the block from cache if
 * possible, and updates cache entries.
 * To avoid data loss, writing data is ALWAYS in sync. This means, this function
 * writes data on MMC regardless of whether it is in cache or not. */
static int mmc_do_write(unsigned int addr, unsigned int count, const void *buf)
{
	char writedata[BLOCKSIZE];
	int r;

	unsigned int blocknr = addr / BLOCKSIZE;

	/* some words about remainder parts and what we have to do in this func
	 * and so on.
	 *
	 * MMC cards write their data blockwise. If we just want to write 1 byte
	 * in a block, we have to write the whole block. This means, we have to
	 * know the content of the other bytes from that block. */

	/* check whether remainder parts are there. */
	if (addr % BLOCKSIZE == 0 && count == BLOCKSIZE) {
		/* no remainder parts. perfect, no copying needed. */
		return mmc_write_block(blocknr, buf);
	} else {
		if ((r = mmc_read_block(blocknr * BLOCKSIZE, BLOCKSIZE,
					writedata)) < 0)
			return r;
		memcpy(writedata + (addr % BLOCKSIZE), buf, count);
		return mmc_write_block(blocknr, writedata);
	}
}

static int mmc_write(struct filedes *fd, const void *buf, unsigned int count)
{
	/* almost the same as mmc_read() */

	unsigned int cc;
	unsigned int j = 0, pos = fd->pos;
	int r;

	while (j < count) {
		/* Three things must be considered for setting cc:
		 * 1. current addr might be behind the beginning of a block
		 * 2. current addr might be at the beginning of a block
		 * 3. current addr might be in the last requested block */
		cc = BLOCKSIZE - pos % BLOCKSIZE;	/* 1 */
		if (!cc)				/* 2 */
			cc = BLOCKSIZE;
		if (pos + cc > fd->pos + count)		/* 3 */
			cc = fd->pos + count - pos;

		if ((r = mmc_do_write(pos, cc, buf + j)) < 0)
			return r;

		pos += cc;
		j   += cc;
	}

	return j;
}

static const struct io_handle mmc_ioh = {
	.open		= def_open,
	.close		= def_close,
	.clone_fd	= def_clone_fd,
	.read		= mmc_read,
	.write		= mmc_write,
	.poll		= def_poll
};

void mmc_init(void)
{
	int i;
	int timeout;

	if (!mmc_card_inserted()) {
		printf("mmc: no card inserted\n");
		return;
	}

	/* define all output pins (= MMC' input pins) as output pins and set
	 * them to high */
	IODIR |= CS | DI | CLK;
	IOSET = CS | DI | CLK;

	/* send some SPI dummy bytes to the card */
	for (i = 0; i < 15; i++)
		sspi_write(0xff, LOW_SPEED_MODE);

	/* GO_IDLE_STATE is used to reset the MMC and to tell it, that we hate
	 * the MMC interface and would like to use SPI */
	for (timeout = TIMEOUT;
	     mmc_command(GO_IDLE_STATE, 0, LOW_SPEED_MODE) != 1 && timeout;
	     timeout--) ;
	IOSET = CS;
	if (!timeout) {
		puts("mmc: timed out.");
		return;
	}

	/* Tell the card to initalize itself */
	for (timeout = TIMEOUT;
	     mmc_command(SEND_OP_COND, 0, LOW_SPEED_MODE) != 0 && timeout;
	     timeout--) ;
	IOSET = CS;
	if (!timeout) {
		puts("mmc: timed out.");
		return;
	}

	/* TODO: read CSD */
	puts("mmc: initalized.");

	devfs_add("mmc", &mmc_ioh, 0);

	ibmpart_scan("mmc", &mmc_ioh);
}
