/* Copyright (c) 2009-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "pf/cpu.h"
#include "page.h"
#include "errno.h"
#include "lib.h"
#include "exec.h"
#include "sched.h"

#define PAGE_USED	(1 << 15)

/* mask of the tid field of a page entry in the page table */
#define PU_TIDMASK	0x0fff

#define PAGE_MASK	~(PAGE_SIZE - 1)

#ifdef HAS_EXTERNAL_RAM
# define RAM_NO		2
#else
# define RAM_NO		1
#endif

extern void __bss_end__;

static short pt1[RAM_SIZE / PAGE_SIZE];

#ifdef HAS_EXTERNAL_RAM
static short pt2[EXTRAM_SIZE / PAGE_SIZE];
#endif

static const struct {
	/* start of this area */
	unsigned long start;

	/* number of pages in that area */
	int pages;

	/* page table */
	short *pt;
} ram_area[RAM_NO] = {
	{ RAM_START, RAM_SIZE / PAGE_SIZE, pt1 },
#ifdef HAS_EXTERNAL_RAM
	{ EXTRAM_START, EXTRAM_SIZE / PAGE_SIZE, pt2 }
#endif
};

void page_init(void)
{
	long i;
	long bsssize = (long) &__bss_end__ - RAM_START;

	/* allocate bss space */
	for (i = 0; i < bsssize; i += PAGE_SIZE)
		ram_area[0].pt[i / PAGE_SIZE] =
			PAGE_USED | PROT_READ | PROT_WRITE;

	/* allocate kernel stack space */
	for (i = ram_area[0].pages - KERN_STACKSIZE; i < ram_area[0].pages; i++)
		ram_area[0].pt[i] = PAGE_USED | PROT_READ | PROT_WRITE;
}

void *page_alloc(int tid, int n, int prot)
{
	int i, j, k;

	for (i = 0; i < RAM_NO; i++) {
		for (j = 0; j + n < ram_area[i].pages; j++) {
			if (ram_area[i].pt[j])
				/* used */
				continue;

			/* check whether there are enough continious free pages */
			for (k = j + 1; k < j + n; k++)
				if (ram_area[i].pt[k])
					break;

			if (k < j + n) {
				/* not enough free pages */
				j = (k - 1);
				continue;
			}

			break;
		}

		if (j + n < ram_area[i].pages)
			break;
	}

	if (i == RAM_NO)
		/* TODO: invent OOM killer */
		kpanic("Out of memory");

	/* reserve pages */
	for (k = j; k < n + j; k++)
		ram_area[i].pt[k] = tid | PAGE_USED | prot;

	return (void *) (ram_area[i].start + j * PAGE_SIZE);
}

int page_free(const void *ptr, int n)
{
	int i, j;
	int start;

	for (i = 0; i < RAM_NO; i++)
		if ((unsigned long) ptr >= ram_area[i].start &&
		    (unsigned long) ptr <  ram_area[i].start + ram_area[i].pages * PAGE_SIZE)
			break;

	if (i == RAM_NO)
		/* TODO: kpanic("Attempted to free page at invalid address"); */
		return -EINVAL;

	start = ((unsigned long) ptr - ram_area[i].start) / PAGE_SIZE;

	for (j = start; j < start + n; j++)
		ram_area[i].pt[j] = 0;

	return 0;
}

void page_free_all(int tid)
{
	int i, j;

	for (i = 0; i < RAM_NO; i++)
		for (j = 0; j < ram_area[i].pages; j++)
			if ((ram_area[i].pt[j] & PU_TIDMASK) == tid)
				ram_area[i].pt[j] = 0;
}

void *sys_mmap(int tid, void *addr, int len, int prot, int flags, int fildes,
	       int off)
{
	/* TODO: support more */
	/* TODO: do that posix conform */

	(void) addr;
	(void) fildes;
	(void) off;

	void *rv;
	int pages;

	if (prot == PROT_NONE ||
	    !(flags & MAP_PRIVATE) ||
	    flags & MAP_FIXED ||
	    !(flags & MAP_ANON))
		return (void *) ENOTSUP;

	pages = blocks(len, PAGE_SIZE);
	rv = page_alloc(tid, pages, prot);
#ifdef CONFIG_USRZERO
	memset(rv, 0, pages * PAGE_SIZE);
#endif
	return rv;
}

int sys_munmap(void *addr, int len)
{
	if (!len)
		return -EINVAL;
	/* TODO: tid checking */
	return page_free(addr, blocks(len, PAGE_SIZE));
}

int is_in_ram(void *addr)
{
	int i;
	unsigned long uaddr = (unsigned long) addr;

	for (i = 0; i < RAM_NO; i++)
		if (ram_area[i].start <= uaddr &&
		    uaddr < ram_area[i].start + ram_area[i].pages * PAGE_SIZE)
			break;

	return i < RAM_NO;
}

int page_get_total(void)
{
	int i;
	int rv = 0;

	for (i = 0; i < RAM_NO; i++)
		rv += ram_area[i].pages;

	return rv;
}

int page_get_pertid(int tid)
{
	int i, j;

	int rv = 0;

	for (i = 0; i < RAM_NO; i++)
		for (j = 0; j < ram_area[i].pages; j++)
			if (ram_area[i].pt[j] & PAGE_USED &&
			    (ram_area[i].pt[j] & PU_TIDMASK) == tid)
				rv++;

	return rv;
}

#ifdef CONFIG_PTRCHECK

int ptr_valid(void *ptr, int tid, int prot, int size)
{
	int ra;
	unsigned long lp = (unsigned long) ptr;
	unsigned long startaddr, endaddr;
	int startpg, endpg;
	int i;

	if (!usrspc_started)
		return 1;

	if (size == 0)
		return 1;

	/* continous area can only be at one ram area */
	for (ra = 0; ra < RAM_NO; ra++)
		if (lp >= ram_area[ra].start &&
		    lp <  ram_area[ra].start + ram_area[ra].pages * PAGE_SIZE)
			break;
	if (ra == RAM_NO)
		return 0;

	if (lp + size > ram_area[ra].start + ram_area[ra].pages * PAGE_SIZE)
		return 0;

	/* TODO:
	if (!procs[tid]->img_replaced)
		tid = tid of parent;*/

	size--;	/* due to <= */
	startaddr = lp & PAGE_MASK;
	endaddr   = lp + size;
	startpg = (startaddr - ram_area[ra].start) / PAGE_SIZE;
	endpg   = (endaddr   - ram_area[ra].start) / PAGE_SIZE;
	for (i = startpg; i <= endpg; i++)
		if ((procs[tid]->img_replaced && (ram_area[ra].pt[i] & PU_TIDMASK) != tid) ||
		    !(ram_area[ra].pt[i] & PAGE_USED) ||
		    !(ram_area[ra].pt[i] & prot))
			return 0;

	return 1;
}

int ptr_valid_string(char *ptr, int tid)
{
	int left;

	while (1) {
		if (!ptr_valid(ptr, tid, PROT_READ, 1))
			return 0;

		left = PAGE_SIZE - ((long) ptr % PAGE_SIZE);
		while (left--) {
			if (!*ptr)
				return 1;
			ptr++;
		}
	}
}

int ptr_valid_strarray(char **ptr, int tid)
{
	int left;

	if ((long) ptr % 4 != 0)
		return 0;

	while (1) {
		if (!ptr_valid(ptr, tid, PROT_READ, 4))
			return 0;

		left = (PAGE_SIZE - ((long) ptr % PAGE_SIZE)) / 4;
		while (left--) {
			if (!*ptr)
				return 1;
			if (!ptr_valid_string(*ptr, tid))
				return 0;
			ptr++;
		}
	}
}

#endif
