/* Copyright (c) 2009-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "bio.h"
#include "errno.h"
#include "isr.h"
#include "lib.h"
#include "page.h"
#include "sched.h"
#include "pf/cpu.h"
#include "asm.h"
#include "syscall.h"
#include "terminal.h"

#ifdef CONFIG_WDT
/* Watchdog Timer */
#include "pf/wdt.h"
#endif

/* timer registers */
#define T0IR	(* (volatile int *) (TIMER0_BASE+0x00))
#define T0TCR	(* (volatile int *) (TIMER0_BASE+0x04))
#define T0TC	(* (volatile int *) (TIMER0_BASE+0x08))
#define T0MCR	(* (volatile int *) (TIMER0_BASE+0x14))
#define T0MR0	(* (volatile int *) (TIMER0_BASE+0x18))

static void timer_init(void);
static int alloc_pid(void);
static int proc_block(struct context *ctx, int tid, int wt_func);
static int proc_wake(int tid, int wt_func);
static void poll_cleanup(int tid);
static void sched_cleanup(int tid);
static int la_get(int minutes);
static unsigned int __optimize uptime_get(void);

/* load average computation:
 *
 * On every sched() issued by the timer, the number of running processes is
 * counted and added to la_period_running. Additionally, la_lastcalc is
 * incremented by one. If la_lastcalc reaches SCHED_LA_TICKS, everything in
 * la_calcs is shiftet one index towards infinity and la_calcs[0] becomes
 * la_period_running. la_period_running and la_lastcalc is reset to 0.
 *
 * Thus, la_calcs stores the average number of running processes *
 * SCHED_LA_TICKS for a period of 60 / SCHED_LA_PERMIN seconds. To get the load
 * average of the last minute, you have to count la_calcs[0:3] and divide it by
 * (4 * SCHED_LA_TICKS). */
static int la_period_running;
static int la_lastcalc;
static unsigned short la_calcs[SCHED_LA_TOTCLC];

struct process *procs[SCHED_MAX_PROC];
int current;

static unsigned int timer_ticks;


/*
 * scheduling and initalization functions
 *
 * (timer_init, alloc_pid, sched_init, sched, sched_timer)
 */


static void timer_init(void)
{
	T0MR0 = (CPU_CLOCK / SCHED_HZ);
	T0MCR = 3;	/* reset, interrupt */
	T0TCR = 1;	/* enable */
}

static int alloc_pid(void)
{
	static int lastpid;
	int ret;
	int i;

retry:
	ret = lastpid++;

	/* check if pid is too high */
	if (ret == SCHED_MAX_PID) {
		lastpid = 2;	/* 0 and 1 are always used */
		goto retry;
	}

	/* check if pid is already used */
	for (i = 0; i < SCHED_MAX_PROC; i++)
		if (procs[i] && procs[i]->pid == ret)
			goto retry;

	return ret;
}

void sched_init(void)
{
	int p;

	/* create idle process running when there is no other process to run. */
	p		= alloc_pid();
	procs[0]	= malloc(sizeof(struct process));
	procs[0]->pid	= p;
	procs[0]->ppid	= 0;
	procs[0]->fds	= 0;
	procs[0]->state	= ST_RUNNING;
	procs[0]->stack	= 0;
	procs[0]->cmdline = (char *) (int) "[idle]";
	procs[0]->img_replaced = 1;
	procs[0]->cputime = 0;
	procs[0]->ctx.psr = PSR_M_usr | PSR_F;	/* no privileged mode needed */
	procs[0]->ctx.pc  = (long) idle;
	sig_reset_handlers(0);

	/* kernel initalization routine */
	p		= alloc_pid();
	procs[1]	= malloc(sizeof(struct process));
	procs[1]->pid	= p;
	procs[1]->ppid	= 0;
	procs[1]->fds	= 0;
	procs[1]->state	= ST_RUNNING;
	procs[1]->stack	= page_alloc(1, SCHED_PROC_SP, PROT_READ | PROT_WRITE);
	procs[1]->cmdline = 0;
	procs[1]->img_replaced = 0;
	procs[1]->cputime = 0;
	procs[1]->ctx.psr = PSR_M_sys | PSR_F | PSR_I;
	procs[1]->ctx.sp  = (long) (procs[1]->stack + SCHED_PROC_SP * PAGE_SIZE);
	procs[1]->ctx.pc  = (long) init2;
	sig_reset_handlers(1);

	/* we are starting in running mode... so indicate that */
	led_indicate_running();

	/* small trick to let first task switch work */
	current = SCHED_MAX_PROC - 1;

#ifdef CONFIG_WDT
	/* initalize watchdog timer */
	wdt_init();
#endif

	/* initalize timer */
	vic_set(TIMER_IRQ, isr_timer);
	timer_init();

	/* cause task switch (call sched_yield syscall) */
	asm ("swi %[sc]" :: [sc] "i" (SC_SCHED_YIELD));
}

void __optimize sched(struct context *ctx, int from_timer)
{
	unsigned int cut;
	int i, next, pc;

	/* find next process to switch to */
	pc = current ? current : 1;
	next = 0;
	i = pc;
	do {
		i++;

		if (i == SCHED_MAX_PROC)
			i = 1;	/* not 0. idle (0) is left out when searching
				   process */

		if (!procs[i])
			continue;

		/* if we are called from timer, compare all timeout_tick of the
		 * polling processes, if timeout expired, wake the process up
		 * and clean its polling information */
		if (from_timer &&
		    procs[i]->state == ST_SLEEPING &&
		    procs[i]->sti.sleeping.wt_func == WT_POLL &&
		    procs[i]->sti.sleeping.timeout_tick != 0)
			if (timer_ticks >= procs[i]->sti.sleeping.timeout_tick) {
				/* wake this up.. its timeout expired */
				procs[i]->state = ST_RUNNING;
				procs[i]->ctx.r0 = 0;
				next = i;
				poll_cleanup(i);
			}

		/* if we found a RUNNING_NEXT process, call it and no other
		 * process */
		if (procs[i]->state == ST_RUNNING_NEXT) {
			procs[i]->state = ST_RUNNING;
			next = i;
			if (!from_timer)
				/* only break if we are not scheduled by the
				 * timer. If the timer causes the schedule we
				 * have to go through the whole list to update
				 * state of timeouts */
				break;
		}

		/* if we found a normal RUNNING process and did not found an
		 * other such process yet, take it as next process. But we still
		 * do not leave the loop as there could be a RUNNING_NEXT
		 * process which has to be called first. */
		if (!next && procs[i]->state == ST_RUNNING)
			next = i;

		/* load average computation */
		if (from_timer && procs[i]->state == ST_RUNNING)
			la_period_running++;
	} while (i != pc);

	if (next == 0 && current != 0)
		led_indicate_idle();
	if (current == 0 && next != 0)
		led_indicate_running();

	if (next != current) {
		/* set got_running for CPU time update when another process is
		 * running */
		cut = uptime_get();
		procs[next]->got_running = cut;

		if (procs[current]) {
			/* update CPU time */
			procs[current]->cputime +=
				cut - procs[current]->got_running;

			/* store context of current process */
			memcpy(&procs[current]->ctx, ctx,
			       sizeof(struct context));
		}

		/* perform the task switch */
		memcpy(ctx, &procs[next]->ctx, sizeof(struct context));
		current = next;
	}
}

extern void sched_timer(struct context *ctx);

void __optimize sched_timer(struct context *ctx)
{
	/* update uptime */
	timer_ticks++;

	sched(ctx, 1);

	/* load average computation */
	if (++la_lastcalc == SCHED_LA_TICKS) {
		memmove(&la_calcs[1], &la_calcs[0],
			(SCHED_LA_TOTCLC - 1) * sizeof(unsigned short));
		la_calcs[0] = la_period_running;
		la_period_running = la_lastcalc = 0;
	}

#ifdef CONFIG_WDT
	/* feed the watchdog timer to let it know that we are still alive */
	wdt_feed();
#endif

	T0IR = 1;	/* reset timer interrupt */
	vic_reset();	/* reset VIC */
}


/*
 * process managment (creation, state managment) functions
 *
 * (sys_fork, proc_block, proc_wake, proc_block_io, proc_wake_io, sys_poll,
 * poll_cleanup)
 */


int sys_fork(struct context *ctx)
{
	int i, j;
	int p;

	/* TODO: implement ENOMEM errors */

	/* search free process */
	for (i = 0; i < SCHED_MAX_PROC; i++)
		if (!procs[i])
			break;
	if (i == SCHED_MAX_PROC)
		return -EAGAIN;

	/* TODO: copy .bss, .data, ... */

	/* setup some things */
	p		= alloc_pid();
	procs[i]	= malloc(sizeof(struct process));
	procs[i]->pid	= p;
	procs[i]->ppid	= procs[current]->pid;
	procs[i]->state	= ST_RUNNING;
	procs[i]->fds	= 0;
	procs[i]->cmdline = strdup(procs[current]->cmdline);
	procs[i]->img_replaced = 0;
	procs[i]->cputime = 0;

	/* clone file descriptors */
	bio_clone_fds(i, current);

	/* copy signal handlers */
	for (j = 0; j < SIGNUM; j++) {
		procs[i]->sa[j].sa_handler = procs[current]->sa[j].sa_handler;
		procs[i]->sa[j].sa_flags = procs[current]->sa[j].sa_flags;
	}

	/* copy stack */
	procs[i]->stack = page_alloc(i, SCHED_PROC_SP, PROT_READ | PROT_WRITE);
	memcpy(procs[i]->stack, procs[current]->stack,
	       SCHED_PROC_SP * PAGE_SIZE);

	/* setup context */
	memcpy(&procs[i]->ctx, ctx, sizeof(struct context));
	procs[i]->ctx.r0 = 0;		/* return value */
	procs[i]->ctx.sp = ctx->sp - (long) procs[current]->stack +
		(long) procs[i]->stack;	/* stack pointer */

	/* add process to appropiate group */
	term_on_fork(procs[current], procs[i], i);

	/* return child's process id to parent */
	return procs[i]->pid;
}

static int proc_block(struct context *ctx, int tid, int wt_func)
{
	procs[tid]->state = ST_SLEEPING;
	procs[tid]->sti.sleeping.wt_func = wt_func;
	sched(ctx, 0);
	return DONT_RETURN;
}

static int proc_wake(int tid, int wt_func)
{
	if (!procs[tid] ||
	    procs[tid]->state != ST_SLEEPING ||
	    procs[tid]->sti.sleeping.wt_func != wt_func)
		/* it either died or someone already waked it up or it is not
		 * sleeping in the right function */
		return 0;

	procs[tid]->state = ST_RUNNING_NEXT;
	procs[tid]->ctx.pc -= 4;	/* thus the swi instruction which caused
					 * it to sleep (and now won't block it
					 * anymore) is called again */

	return 1;
}

int proc_block_io(struct context *ctx, int tid, int wt_func, int event,
		  struct filedes *fd)
{
	procs[tid]->sti.sleeping.iowait_info = malloc(sizeof(struct iowait_info));
	procs[tid]->sti.sleeping.iowait_info->event = event;
	procs[tid]->sti.sleeping.iowait_info->fd = fd;
	return proc_block(ctx, tid, wt_func);
}

void proc_wake_io(struct context *ctx, int event, const struct io_handle *ioh)
{
	int i, j;
	int v = 0;
	struct iowait_info *iowi;

	/* TODO: repolling */

	for (i = 0; i < SCHED_MAX_PROC && !v; i++) {
		if (!procs[i] || procs[i]->state != ST_SLEEPING)
			continue;

		if ((procs[i]->sti.sleeping.wt_func == WT_READ ||
		     procs[i]->sti.sleeping.wt_func == WT_WRITE) &&
		    procs[i]->sti.sleeping.iowait_info->fd->ioh == ioh &&
		    procs[i]->sti.sleeping.iowait_info->event & event) {
			/* found process blocked in read or write and waiting
			 * for exactly this event to happen */
			free(procs[i]->sti.sleeping.iowait_info);
			proc_wake(i, procs[i]->sti.sleeping.wt_func);
			v++;
		}

		if (procs[i]->sti.sleeping.wt_func == WT_POLL)
			/* have to go through the process iowait_info list and
			 * look whether there are matching ones */
			for (iowi = procs[i]->sti.sleeping.iowait_info, j = 0;
			     iowi;
			     iowi = iowi->next, j++)
				if (iowi->fd->ioh == ioh &&
				    iowi->event & event) {
					/* matched... wake this up and end the
					 * poll by cleaning up */
					procs[i]->state = ST_RUNNING_NEXT;
					procs[i]->ctx.r0 = 1;
					procs[i]->sti.sleeping.pollfd[j].revents =
						iowi->event & event;
					poll_cleanup(i);
					v++;
					break;
				}
	}

	if (v)
		/* only switch task if we acutally waked one up */
		sched(ctx, 0);
}

int sys_poll(struct context *ctx, struct pollfd *fds, unsigned int nfds,
	     int timeout)
{
	unsigned int i;
	struct filedes *fd;
	int r = 0;
	int e;
	struct iowait_info *newinf, *inf = 0;

	/* poll all given filedescriptors */
	for (i = 0; i < nfds; i++) {
		fds[i].revents = 0;

		if (!bio_get_fdinfo(current, fds[i].fd, &fd)) {
			fds[i].revents = POLLNVAL;
			r++;
			continue;
		}

		e = fd->ioh->poll(fd, fds[i].events);
		if (e) {
			fds[i].revents = e;
			r++;
		}
	}

	/* return if either an awaited event already occured or timeout is 0 */
	if (r || !timeout)
		return r;

	/* can't return yet */

	/* build up information struct */
	if (timeout == -1)
		procs[current]->sti.sleeping.timeout_tick = 0;
	else {
		procs[current]->sti.sleeping.timeout_tick =
			timer_ticks + SCHED_HZ * timeout / 1000;
		if ((SCHED_HZ * timeout) % 1000 != 0)
			/* POSIX about poll(): "Implementations may place
			 * limitations on the granularity of timeout intervals.
			 * If the requested timeout interval requires a finer
			 * granularity than the implementation supports, the
			 * actual timeout interval shall be rounded up to the
			 * next supported value."
			 * This way, we also ensure that timeout_tick is not set
			 * to 0 (= no timeout) if the caller wants a timeout,
			 * because if timeout=0, poll() returns immediately, and
			 * if timeout > 0, but < 1000/SCHED_HZ, timeout_tick is
			 * at least timer_ticks + 1. */
			procs[current]->sti.sleeping.timeout_tick++;
	}
	procs[current]->sti.sleeping.pollfd = fds;
	for (i = 0; i < nfds; i++) {
		newinf = malloc(sizeof(struct iowait_info));
		if (i == 0)
			procs[current]->sti.sleeping.iowait_info = newinf;
		else
			inf->next = newinf;
		inf = newinf;
		
		inf->event = fds[i].events;
		bio_get_fdinfo(current, fds[i].fd, &inf->fd);

		inf->next = 0;
	}
	if (!nfds)
		procs[current]->sti.sleeping.iowait_info = 0;

	/* block
	 *
	 * poll must not be waked up with proc_wake, as it just executes the swi
	 * poll syscall again, and thus resets our timeout. */
	return proc_block(ctx, current, WT_POLL);
}

static void poll_cleanup(int tid)
{
	struct iowait_info *iowi = procs[tid]->sti.sleeping.iowait_info;

	while (iowi) {
		free(iowi);
		iowi = iowi->next;
	}
}


/*
 * process removal functions
 *
 * (sys__exit, sched_cleanup, sys_waitpid)
 */


void sys__exit(struct context *ctx, int tid, int status)
{
	int parent;

	/* TODO: be POSIX conform */

	if (tid == 1)
		/* init is the parent of all processes. it must survive */
		kpanic("init called _exit(%i)", status);

	page_free_all(tid);

	bio_close_all(tid);

#ifdef CONFIG_SYMTAB
	free(procs[tid]->elfsymstrtab);
	free(procs[tid]->elfsymtab);
#endif

	procs[tid]->state = ST_ZOMBIE;
	procs[tid]->sti.zombie.exit_status = status;

	/* search parent */
	for (parent = 0; parent < SCHED_MAX_PROC; parent++)
		if (procs[parent] &&
		    procs[parent]->pid == procs[tid]->ppid)
			break;
	if (parent == SCHED_MAX_PROC)
		/* this is a really big fail and can't happen if the scheduler
		 * code works */
		kpanic("no parent found");

	/* wake up our parent if he is sleeping */
	if (!proc_wake(parent, WT_WAITPID)) {
		/* parent was not waiting. if it has the SA_NOCLDWAIT flag set
		 * or set SIGCHLD to SIG_IGN, we can immediately delete the
		 * process instead of transforming it to a zombie. */
		if (procs[parent]->sa[SIGCHLD].sa_handler == (void *) SIG_IGN ||
		    procs[parent]->sa[SIGCHLD].sa_flags & SA_NOCLDWAIT)
			sched_cleanup(tid);
		else
			/* send SIGCHLD */
			sig_emit(ctx, parent, SIGCHLD);
	}

	/* remove process from its group and sent SIGHUP if needed */
	term_on_exit(ctx, procs[tid]);

	sched(ctx, 0);
}

/* used to transfer a zombie process to an actually dead process */
static void sched_cleanup(int tid)
{
	int i;

	/* set init as father of all our child processes */
	for (i = 0; i < SCHED_MAX_PROC; i++)
		if (procs[i] && procs[i]->ppid == procs[tid]->pid)
			procs[i]->ppid = 1;

	/* remove us */
	free(procs[tid]->cmdline);
	free(procs[tid]);
	procs[tid] = 0;
}

int sys_waitpid(struct context *ctx, int pid, int *stat_loc, int options)
{
	int i;

	/* TODO: be POSIX conform (return more -ECHILD) and more options and
	 * more pid values */

	/* check for zombie processes matching our criteria */
	for (i = 0; i < SCHED_MAX_PROC; i++)
		if (procs[i] &&
		    procs[i]->state == ST_ZOMBIE &&
		    (pid == -1 || procs[i]->pid == pid) &&
		    procs[i]->ppid == procs[current]->pid) {
			/* found one */
			if (stat_loc)
				*stat_loc = procs[i]->sti.zombie.exit_status & 0xff;
			sched_cleanup(i);
			return procs[i]->pid;
		}

	/* if the above loop does not return, no process was found */

	if (options & WNOHANG)
		return -ECHILD;

	/* set us to waiting status */
	return proc_block(ctx, current, WT_WAITPID);
}


/*
 * signal handling functions
 *
 * (sig_defaults, sig_reset_handlers, sys_sigrestorectx, sig_emit, sys_kill,
 * sys_sigaction, sigact_kill, sigact_dumpkill, sigact_stop, sigact_continue)
 */


/* handlers for SIG_DFL */
static void sigact_kill(struct context *ctx, int tid, int sig);
static void sigact_dumpkill(struct context *ctx, int tid, int sig);
static void sigact_stop(struct context *ctx, int tid, int sig);
static void sigact_continue(struct context *ctx, int tid, int sig);

/* default signal configuration */
static const struct {
	/* default action setting */
	void *dfl;

	/* handler for SIG_DFL */
	void (*hdl)(struct context *ctx, int tid, int sig);
} sig_defaults[SIGNUM] = {
	[0]		= { SIG_IGN, 0 },		/* invalid */
	[SIGHUP]	= { SIG_DFL, sigact_kill },
	[SIGINT]	= { SIG_DFL, sigact_kill },
	[SIGQUIT]	= { SIG_DFL, sigact_dumpkill },
	[SIGBUS]	= { SIG_DFL, sigact_dumpkill },
	[SIGCHLD]	= { SIG_IGN, 0 },
	[SIGABRT]	= { SIG_DFL, sigact_dumpkill },
	[SIGCONT]	= { SIG_DFL, sigact_continue },
	[SIGFPE]	= { SIG_DFL, sigact_dumpkill },
	[SIGKILL]	= { SIG_DFL, sigact_kill },
	[SIGILL]	= { SIG_DFL, sigact_dumpkill },
	[SIGPIPE]	= { SIG_DFL, sigact_kill },
	[SIGPOLL]	= { SIG_DFL, sigact_kill },
	[SIGPROF]	= { SIG_DFL, sigact_kill },
	[SIGALRM]	= { SIG_DFL, sigact_kill },
	[SIGTERM]	= { SIG_DFL, sigact_kill },
	[SIGSEGV]	= { SIG_DFL, sigact_dumpkill },
	[SIGSTOP]	= { SIG_DFL, sigact_stop },
	[SIGSYS]	= { SIG_DFL, sigact_dumpkill },
	[SIGTRAP]	= { SIG_DFL, sigact_dumpkill },
	[SIGTSTP]	= { SIG_DFL, sigact_stop },
	[SIGTTIN]	= { SIG_DFL, sigact_stop },
	[SIGTTOU]	= { SIG_DFL, sigact_stop },
	[SIGURG]	= { SIG_IGN, 0 },
	[SIGUSR1]	= { SIG_DFL, sigact_kill },
	[SIGUSR2]	= { SIG_DFL, sigact_kill },
	[SIGUSR3]	= { SIG_DFL, sigact_kill },
	[SIGUSR4]	= { SIG_DFL, sigact_kill },
	[SIGUSR5]	= { SIG_DFL, sigact_kill },
	[SIGVTALRM]	= { SIG_DFL, sigact_kill },
	[SIGXCPU]	= { SIG_DFL, sigact_dumpkill },
	[SIGXFSZ]	= { SIG_DFL, sigact_dumpkill }
};

void sig_reset_handlers(int tid)
{
	int i;

	for (i = 0; i < SIGNUM; i++) {
		procs[tid]->sa[i].sa_handler = (void *) sig_defaults[i].dfl;
		procs[tid]->sa[i].sa_flags = 0;
	}
}

/* restores the context after returning from the signal handler */
void sys_sigrestorectx(void)
{
	if (!procs[current]->sigctx) {
		/* process is not in signal. thus, this is an invalid syscall */
		sig_emit(sc_ctx, current, SIGSYS);
		return;
	}

	procs[current]->state = procs[current]->sigctx->state;
	memcpy(sc_ctx, &procs[current]->sigctx->ctx, sizeof(struct context));

	if (procs[current]->state == ST_SLEEPING) {
		if (procs[current]->sa[procs[current]->sigctx->signum].sa_flags
		    & SA_RESTART)
			sc_ctx->pc -= 4;	/* restart syscall */
		else
			sc_ctx->r0 = -EINTR;	/* end syscall with EINTR */
		procs[current]->state = ST_RUNNING_NEXT;
	}

	free(procs[current]->sigctx);
	procs[current]->sigctx = 0;

	/* TODO: let kill return 0 on self-killing */

	sched(sc_ctx, 0);
}

/* this is NOT sys_kill() (too few checks, etc.) */
void sig_emit(struct context *ctx, int tid, int sig)
{
	struct context *newctx;

	/* TODO: after SIG_IGN handling */
	if (procs[tid]->sa[sig].sa_flags & SA_RESETHAND) {
		procs[tid]->sa[sig].sa_handler = (void *) sig_defaults[sig].dfl;
		procs[tid]->sa[sig].sa_flags = 0;
	}

	if (procs[tid]->sa[sig].sa_handler == SIG_IGN)
		return;

	if (procs[tid]->sa[sig].sa_handler == SIG_DFL) {
		sig_defaults[sig].hdl(ctx, tid, sig);
		return;
	}

	/* signal is to be handled by userspace */

	if (procs[tid]->sigctx)
		/* already in signal handling function...
		 * TODO: have signal queue */
		return;

	if (procs[tid]->state == ST_ZOMBIE || procs[tid]->state == ST_STOPPED)
		/* these processes can't handle a signal
		 * TODO: what do we have to do with them? */
		return;

	/* back up current context */
	procs[tid]->sigctx = malloc(sizeof(struct sigctx));
	procs[tid]->sigctx->state = procs[tid]->state;
	procs[tid]->sigctx->signum = sig;
	if (tid == current)
		/* process kills itself (raise()). We have to back up the
		 * context passed to this function instead of the old context in
		 * the process list */
		memcpy(&procs[tid]->sigctx->ctx, ctx, sizeof(struct context));
	else
		memcpy(&procs[tid]->sigctx->ctx, &procs[tid]->ctx,
		       sizeof(struct context));
	/* TODO: back up I/O wait information */

	/* create new context */
	procs[tid]->state = ST_RUNNING_NEXT;
	if (tid == current)
		/* if we call sched(), it writes the passed ctx to the ctx of
		 * the current process. Thus we have to modify our parameter ctx
		 * if the process kills itself. */
		newctx = ctx;
	else
		newctx = &procs[tid]->ctx;
	memset(newctx, 0, sizeof(struct context));
	newctx->psr = PSR_M_usr | PSR_F;
	newctx->r0  = sig;
	newctx->sp  = procs[tid]->sigctx->ctx.sp;	/* TODO: SA_ONSTACK */
	newctx->pc  = (long) procs[tid]->sa[sig].sa_handler;
	/* this is a small trick to let the syscall handler restore its context
	 * afterwards. This function (sig_lrfunc()) invokes the
	 * sys_sigrestorectx() syscall. The function is set as return link so
	 * the signal handler calls it when returning. */
	newctx->lr  = (long) sig_lrfunc;

	sched(ctx, 0);
}

/* wrapper around sig_emit(), secure enough for userspace */
int sys_kill(int pid, int sig)
{
	int tid;

	/* start tid searching at tid 1. This makes it impossible to kill idle
	 * task */
	for (tid = 1; tid < SCHED_MAX_PROC; tid++)
		if (procs[tid] && procs[tid]->pid == pid)
			break;
	if (tid == SCHED_MAX_PROC)
		return -ESRCH;

	/* check of sig parameter must be done after checking for error
	 * conditions, because POSIX says "If sig is 0 (the null signal), error
	 * checking is performed but no signal is actually sent." */
	if (!sig)
		return 0;

	if (sig < 0 || sig >= SIGNUM)
		return -EINVAL;

	sig_emit(sc_ctx, tid, sig);

	/* TODO: not reliable */
	return (tid == current) ? DONT_RETURN : 0;
}

int sys_sigaction(int sig, const struct sigaction *act, struct sigaction *oact)
{
	if (sig <= 0 || sig >= SIGNUM)
		return -EINVAL;

	if (oact)
		memcpy(oact, &procs[current]->sa[sig], sizeof(struct sigaction));

	if (!act)
		return 0;

	if (sig == SIGKILL || sig == SIGSTOP)
		/* these signals are not catchable by userspace */
		return -EINVAL;

	memcpy(&procs[current]->sa[sig], act, sizeof(struct sigaction));
	return 0;
}

/* POSIX: "Abnormal termination of the process." */
static void sigact_kill(struct context *ctx, int tid, int sig)
{
	if (tid == 0 || tid == 1)
		kpanic("Attempted to kill init or idle process");

	/* TODO: be POSIX conform */
	sys__exit(ctx, tid, sig << 8);
}

/* POSIX: "Abnormal termination of the process with additional actions." */
static void sigact_dumpkill(struct context *ctx, int tid, int sig)
{
	/* dumps registers, callstack, ... */
	exception(ctx, tid, sig);

	sigact_kill(ctx, tid, sig);
}

/* POSIX: "Stop the process." */
static void sigact_stop(struct context *ctx, int tid, int sig)
{
	(void) sig;

	if (procs[tid]->state == ST_STOPPED)
		return;

	procs[tid]->old_state = procs[tid]->state;
	procs[tid]->state = ST_STOPPED;

	sched(ctx, 0);
}

/* POSIX: "Continue the process, if it is stopped; otherwise, ignore the
 * signal." */
static void sigact_continue(struct context *ctx, int tid, int sig)
{
	(void) sig;

	if (procs[tid]->state == ST_STOPPED) {
		procs[tid]->state = procs[tid]->old_state;
		sched(ctx, 0);
	}
}


/*
 * statistic retrieval functions
 *
 * (la_get, uptime_get, sys_sched_info, sys_sched_list)
 */


static int la_get(int minutes)
{
	int i;
	int sum = 0;

	for (i = 0; i < minutes * SCHED_LA_PERMIN; i++)
		sum += la_calcs[i];

	return 1000 * sum / SCHED_LA_TICKS / minutes / SCHED_LA_PERMIN;
}

static unsigned int __optimize uptime_get(void)
{
	unsigned int rv;

	rv = 1000 * timer_ticks / SCHED_HZ;
	rv += 1000 * T0TC / CPU_CLOCK;

	return rv;
}

int sys_sched_info(struct sched_info *si)
{
	si->hz		= SCHED_HZ;
	si->procs	= SCHED_MAX_PROC;
	si->memtotal	= page_get_total();
	si->la1		= la_get(1);
	si->la5		= la_get(5);
	si->la15	= la_get(15);
	si->uptime	= uptime_get();

	return 0;
}

int sys_sched_list(struct sched_procinfo *spi, int n)
{
	int i, j;

	for (i = 0, j = 0; i < SCHED_MAX_PROC && j < n; i++)
		if (procs[i]) {
			switch (procs[i]->state) {
			case ST_ZOMBIE:
				spi[j].state = STATE_ZOMBIE;
				break;
			case ST_RUNNING:
			case ST_RUNNING_NEXT:
				spi[j].state = STATE_RUNNING;
				break;
			case ST_STOPPED:
				spi[j].state = STATE_STOPPED;
				break;
			case ST_SLEEPING:
				spi[j].state = STATE_SLEEPING;
				break;
			}

			spi[j].pid	= procs[i]->pid;
			spi[j].ppid	= procs[i]->ppid;
			spi[j].memsz	= page_get_pertid(i);
			spi[j].cmd	= procs[i]->cmdline;
			spi[j].cputime	= procs[i]->cputime;

			j++;
		}

	return j;
}
