/* Copyright (c) 2009-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "bio.h"
#include "exec.h"
#include "sched.h"
#include "syscall.h"
#include "page.h"
#include "errno.h"
#include "terminal.h"
#include "lib.h"

extern void syscall(struct context *ctx);

struct context *sc_ctx;

void syscall(struct context *ctx)
{
	int rv = DONT_RETURN;

	/* read out 24 bit immediate field of the swi instruction */
	int syscall_nr = *((long *) (ctx->pc - 4)) & 0x00ffffff;

	/* set up sc_ctx variable for syscall handlers needing direct access to
	 * the context */
	sc_ctx = ctx;

	switch (syscall_nr) {

			/* process control funtions */
	case SC_SCHED_YIELD:
		sched(ctx, 0);
		break;
	case SC_FORK:
		rv = sys_fork(ctx);
		break;
	case SC__EXIT:
		sys__exit(ctx, current, ctx->r0);
		break;
	case SC_EXECVE:
		if (!ptr_valid_string((char *) ctx->r0, current) ||
		    !ptr_valid_strarray((char **) ctx->r1, current) ||
		    !ptr_valid_strarray((char **) ctx->r2, current))
			rv = -EFAULT;
		else
			rv = elf_exec(ctx, (const char *) ctx->r0,
				      (char **) ctx->r1, (char **) ctx->r2);
		break;
	case SC_WAITPID:
		if (ctx->r1 && !ptr_valid((void *) ctx->r1, current, PROT_WRITE,
					  sizeof(int)))
			rv = -EFAULT;
		else
			rv = sys_waitpid(ctx, ctx->r0, (int *) ctx->r1,
					 ctx->r2);
		break;
	case SC_POLL:
		if (!ptr_valid((void *) ctx->r0, current,
			       PROT_READ | PROT_WRITE,
			       sizeof(struct pollfd) * ctx->r1))
			rv = -EFAULT;
		else
			rv = sys_poll(ctx, (struct pollfd *) ctx->r0, ctx->r1,
				      ctx->r2);
		break;

			/* signal functions */
	case SC_KILL:
		rv = sys_kill(ctx->r0, ctx->r1);
		break;
	case SC_SIGRESTORECTX:
		sys_sigrestorectx();
		break;
	case SC_SIGACTION:
		if ((ctx->r1 && !ptr_valid((void *) ctx->r1, current, PROT_READ,
			       sizeof(struct sigaction))) ||
		    (ctx->r2 && !ptr_valid((void *) ctx->r2, current, PROT_WRITE,
			       sizeof(struct sigaction))))
			rv = -EFAULT;
		else
			rv = sys_sigaction(ctx->r0,
					   (const struct sigaction *) ctx->r1,
					   (struct sigaction *) ctx->r2);
		break;

			/* process info / config functions */
	case SC_GETPID:
		rv = procs[current]->pid;
		break;
	case SC_GETPPID:
		rv = procs[current]->ppid;
		break;
	case SC_SETSID:
		rv = sys_setsid(ctx);
		break;

			/* memory management functions */
	case SC_MMAP:
		rv = (int) sys_mmap(current, (void *) ctx->r0, ctx->r1, ctx->r2,
				    ctx->r3, ctx->r4, ctx->r5);
		break;
	case SC_MUNMAP:
		rv = sys_munmap((void *) ctx->r0, ctx->r1);
		break;

			/* input/output functions */
	case SC_OPEN:
		if (!ptr_valid_string((char *) ctx->r0, current))
			rv = -EFAULT;
		else
			rv = bio_open(current, (const char *) ctx->r0, ctx->r1, 0);
		break;
	case SC_CLOSE:
		rv = bio_close(current, ctx->r0);
		break;
	case SC_READ:
		if (!ptr_valid((void *) ctx->r1, current, PROT_WRITE, ctx->r2))
			rv = -EFAULT;
		else
			rv = bio_read(current, ctx->r0, (void *) ctx->r1,
				      ctx->r2);
		break;
	case SC_WRITE:
		if (!ptr_valid((void *) ctx->r1, current, PROT_READ, ctx->r2))
			rv = -EFAULT;
		else
			rv = bio_write(current, ctx->r0, (const void *) ctx->r1,
				       ctx->r2);
		break;

			/* system info / config functions */
	case SC_SCHED_INFO:
		if (!ptr_valid((void *) ctx->r0, current, PROT_WRITE,
			       sizeof(struct sched_info)))
			rv = -EFAULT;
		else
			rv = sys_sched_info((struct sched_info *) ctx->r0);
		break;
	case SC_SCHED_LIST:
		if (!ptr_valid((void *) ctx->r0, current, PROT_WRITE,
			       sizeof(struct sched_procinfo) * ctx->r1))
			rv = -EFAULT;
		else
			rv = sys_sched_list((struct sched_procinfo *) ctx->r0,
					    ctx->r1);
		break;

	default:
		sig_emit(ctx, current, SIGSYS);
		break;
	}

	if (rv != DONT_RETURN)
		ctx->r0 = rv;

#ifdef CONFIG_PTRCHECK_SEGV
	if (rv == -EFAULT) {
		printf("fault in syscall %o\n", syscall_nr);
		sig_emit(ctx, current, SIGSEGV);
	}
#endif
}
