/* Copyright (c) 2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

/* This file provides generic terminal functions to be used by hardware-specific
 * terminal implementations such as the UART.
 * See uart.c for an example on how to use this API. */

/* Additionally, this file handles session / process group managment, altough it
 * is not terminal specific, but only used by terminal functions. */

/* TODO: be more POSIX conform (especially non-canonical mode, ...) */

#include "bio.h"
#include "dev.h"
#include "errno.h"
#include "lib.h"
#include "sched.h"
#include "terminal.h"

struct term_info {
	/* next terminal in list */
	struct term_info *next;

	/* putchar function as passed on term_init */
	void (*putchar)(char c);

	/* handle for devfs */
	struct io_handle ioh;

	/* receive fifo */
	char recv_fifo[CONFIG_TERM_RECV_FIFO];
	unsigned int recv_fifo_used;
};

static struct term_info *til;

/* linked list around struct process */
struct procent {
	struct procent *next;
	struct process *p;
	int tid;	/* task id for sig_emit */
};

struct pgroup {
	/* next process group in list */
	struct pgroup *next;

	/* process group id */
	int pgid;

	/* linked list of processes in this group */
	struct procent *procs;

	/* group leader, or 0 if the group is orphaned */
	struct process *leader;
};

struct session {
	/* next session in list */
	struct session *next;

	/* terminal associated with this session, or 0 if the session does not
	 * have a controlling terminal */
	struct term_info *term;

	/* session id */
	int sid;

	/* linked list of process groups in this session */
	struct pgroup *pgrps;

	/* foreground group or 0 if there is no foreground group */
	struct pgroup *fgrp;
};

static struct session *sessions;

static void term_send_sig(struct context *ctx, struct term_info *term, int sig);

/* returns count if fifo has data to read (if \n was transmitted) */
static unsigned int term_fifo_has_data(struct term_info *ti)
{
	unsigned int i;

	for (i = 0; i < ti->recv_fifo_used && ti->recv_fifo[i] != '\n'; i++) ;

	if (i == ti->recv_fifo_used)
		return 0;

	return i + 1;	/* include \r */
}

static int term_read(struct filedes *fd, void *buf, unsigned int count)
{
	(void) fd;

	unsigned int r;
	struct term_info *ti = fd->ud;

	r = min(count, term_fifo_has_data(ti));

	/* copy data */
	memcpy(buf, ti->recv_fifo, r);

	if (ti->recv_fifo_used > r)
		/* shift fifo */
		memmove(ti->recv_fifo, ti->recv_fifo + r, ti->recv_fifo_used - r);

	ti->recv_fifo_used -= r;
	return r;
}

static int term_write(struct filedes *fd, const void *buf, unsigned int count)
{
	(void) fd;

	unsigned int i;
	struct term_info *ti = fd->ud;
	const char *cb = buf;

	for (i = 0; i < count; i++)
		ti->putchar(cb[i]);

	return count;
}

static int term_poll(struct filedes *fd, int event)
{
	(void) fd;

	int r;
	struct term_info *ti = fd->ud;

	r = POLLWRNORM;

	if (event & POLLRDNORM && term_fifo_has_data(ti))
		r |= POLLRDNORM;

	return r & event;
}

static void term_clone_fd(struct filedes *dst, struct filedes *src)
{
	dst->ud = src->ud;
}

static int term_open(struct filedes *fd, const char *path)
{
	(void) path;

	struct session *s;
	struct pgroup *g;
	struct procent *p;
	struct term_info *ti = fd->ud;

	if (fd->flags & O_NOCTTY)
		return 0;

	/* check whether terminal is already assigned to a session */
	for (s = sessions; s; s = s->next)
		if (s->term == ti)
			return 0;

	/* check in which group the process is, if any */
	for (s = sessions; s; s = s->next)
		for (g = sessions->pgrps; g; g = g->next)
			for (p = g->procs; p; p = p->next)
				if (p->p == fd->owner)
					goto out; /* to avoid three break;s */
out:
	if (!s)
		/* nothing to do, process has no group */
		return 0;

	/* check whether the process is the controlling process of the
	 * foreground group of an unassigned session */
	if (g->leader == p->p && s->fgrp == g && !s->term)
		/* assign terminal to session */
		s->term = ti;

	return 0;
}

struct term_info *term_init(const char *ttyname, void (*putchar)(char))
{
	struct term_info *t;
	struct term_info *nti = malloc(sizeof(struct term_info));

	if (!til)
		til = nti;
	else {
		for (t = til; t->next; t = t->next) ;
		t->next = nti;
	}

	nti->recv_fifo_used = 0;
	nti->putchar = putchar;
	nti->next = 0;

	nti->ioh.open = term_open;
	nti->ioh.close = def_close;
	nti->ioh.clone_fd = term_clone_fd;
	nti->ioh.read = term_read;
	nti->ioh.write = term_write;
	nti->ioh.poll = term_poll;

	devfs_add(ttyname, &nti->ioh, nti);

	return nti;
}

void term_newc(struct context *ctx, struct term_info *ti, char c)
{
	if (c == 3) {
		term_send_sig(ctx, ti, SIGINT);
		return;
	}

	if (ti->recv_fifo_used >= CONFIG_TERM_RECV_FIFO - 1)
		/* fifo too full
		 * TODO: do something */
		return;

	/* serial end-of-line: '\r',
	 * UNIX   end-of-line: '\n' */
	if (c == '\r')
		c = '\n';

	/* add the char to the FIFO, or remove the last if it was
	 * backspace */
	if (c == '\b') {
		if (ti->recv_fifo_used)
			ti->recv_fifo_used--;
		else
			return;
	} else
		ti->recv_fifo[ti->recv_fifo_used++] = c;

	ti->putchar(c);

	/* hide last-typed char if c is backspace */
	if (c == '\b') {
		ti->putchar(' ');
		ti->putchar('\b');
	}

	if (term_fifo_has_data(ti))
		proc_wake_io(ctx, POLLRDNORM, &ti->ioh);
}

/* send signal to all processes of the given group */
static void group_send_sig(struct context *ctx, struct pgroup *grp, int sig)
{
	struct procent *p;

	for (p = grp->procs; p; p = p->next)
		sig_emit(ctx, p->tid, sig);
}

/* send signal to foreground process of the session associated with the given
 * terminal (if any) */
static void term_send_sig(struct context *ctx, struct term_info *term, int sig)
{
	struct session *s;

	for (s = sessions; s; s = s->next)
		if (s->term == term && s->fgrp) {
			group_send_sig(ctx, s->fgrp, sig);
			break;	/* only one session per terminal */
		}
}

/* remove given process from the given group which is in the given sesssion. */
static void group_remove_proc(struct context *ctx, struct session *ses,
			      struct pgroup *grp, struct procent *proc)
{
	struct procent *p;
	struct pgroup *g;
	struct session *s;
	struct process *pst = proc->p;
	int wasfg = 0;

	/* unlink proc in the list of group members */
	if (grp->procs == proc)
		grp->procs = proc->next;
	else {
		for (p = grp->procs; p->next != proc; p = p->next) ;
		p->next = proc->next;
	}

	free(proc);

	/* check whether process was last in group */
	if (grp->procs == 0) {
		/* unlink group in list of process groups */
		if (ses->pgrps == grp)
			ses->pgrps = grp->next;
		else {
			for (g = ses->pgrps; g->next != grp; g = g->next) ;
			g->next = grp->next;
		}

		free(grp);

		/* check whether group was last in session */
		if (ses->pgrps == 0) {
			/* unlink session in list of sessions */
			if (sessions == ses)
				sessions = ses->next;
			else {
				for (s = sessions; s->next != ses; s = s->next) ;
				s->next = ses->next;
			}

			free(ses);
		} else
			/* if not, check whether it was the foreground group */
			if (ses->fgrp == grp) {
				ses->fgrp = 0;
				wasfg = 1;
			}
	} else
		/* if not, check whether we just orphaned the group */
		if (grp->leader == pst) {
			grp->leader = 0;
			if (wasfg)
				group_send_sig(ctx, grp, SIGHUP);
		}
}

int sys_setsid(struct context *ctx)
{
	struct session *s, *ns;
	struct pgroup *g;
	struct procent *p;

	/* check in which group the process is, if any */
	for (s = sessions; s; s = s->next)
		for (g = sessions->pgrps; g; g = g->next)
			for (p = g->procs; p; p = p->next)
				if (p->p == procs[current])
					goto out; /* to avoid three break;s */
out:
	if (s) {
		/* above loop found matching group. */

		if (g->leader == procs[current])
			/* process is already a group leader */
			return -EPERM;

		/* remove process from group */
		group_remove_proc(ctx, s, g, p);
	}

	/* create new session with new foreground group with this process as
	 * leader. */
	ns = malloc(sizeof(struct session));
	if (!sessions)
		sessions = ns;
	else {
		for (s = sessions; s->next; s = s->next) ;
		s->next = ns;
	}
	ns->next = 0;
	ns->term = 0;
	ns->sid = procs[current]->pid;
	ns->pgrps = malloc(sizeof(struct pgroup));
	ns->pgrps->next = 0;
	ns->pgrps->pgid = procs[current]->pid;
	ns->pgrps->procs = malloc(sizeof(struct procent));
	ns->pgrps->procs->next = 0;
	ns->pgrps->procs->p = procs[current];
	ns->pgrps->procs->tid = current;
	ns->pgrps->leader = procs[current];
	ns->fgrp = ns->pgrps;

	return procs[current]->pid;
}

/* to be called by fork() to add the new process to the group in which the
 * parent process is (if any) */
void term_on_fork(struct process *old, struct process *new, int newtid)
{
	struct session *s;
	struct pgroup *g;
	struct procent *p, *np;

	/* check in which group the old process is, if any */
	for (s = sessions; s; s = s->next)
		for (g = sessions->pgrps; g; g = g->next)
			for (p = g->procs; p; p = p->next)
				if (p->p == old)
					goto out; /* to avoid three break;s */
out:
	if (!s)
		/* nothing to do, old process is in no group */
		return;

	/* add the new process to the group */
	np = malloc(sizeof(struct procent));
	np->next = 0;
	np->p = new;
	np->tid = newtid;
	for (p = g->procs; p->next; p = p->next) ;
	p->next = np;
}

/* to be called by _exit() to remove the exitted process from its process group
 * (if any) */
void term_on_exit(struct context *ctx, struct process *proc)
{
	struct session *s;
	struct pgroup *g;
	struct procent *p;

	/* check in which group the old process is, if any */
	for (s = sessions; s; s = s->next)
		for (g = sessions->pgrps; g; g = g->next)
			for (p = g->procs; p; p = p->next)
				if (p->p == proc)
					goto out; /* to avoid three break;s */
out:
	if (!s)
		/* nothing to do, old process is in no group */
		return;

	group_remove_proc(ctx, s, g, p);
}
