/* Copyright (c) 2009-2010, Alexander Graf <alex@wortkreuzung.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright notice,
 *     this list of conditions and the following disclaimer in the documentation
 *     and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE. */

#include "devices.h"
#include "pf/cpu.h"
#include "terminal.h"

/* registers */
#define DLL	(*(volatile unsigned char *)	(UART0_BASE+0x0 ))
#define DLM	(*(volatile unsigned char *)	(UART0_BASE+0x4 ))
#define THR	(*(volatile char *)		(UART0_BASE+0x0 ))
#define RBR	(*(volatile char *)		(UART0_BASE+0x0 ))
#define IER	(*(volatile int *)		(UART0_BASE+0x4 ))
#define FCR	(*(volatile unsigned char *)	(UART0_BASE+0x8 ))
#define LCR	(*(volatile unsigned char *)	(UART0_BASE+0xc ))
#define LSR	(*(volatile unsigned char *)	(UART0_BASE+0x14))

/* flags for fifo control register (fcr) */
#define FIFO_ENABLE	(1 << 0)
#define RX_RESET	(1 << 1)
#define TX_RESET	(1 << 2)

/* flags for line control register (lcr) */
#define DLAB	(1 << 7)
#define BPB(x)	(x - 5)

/* flags for line status register (lsr) */
#define RDR	(1 << 0)	/* Receiver Data Ready */
#define THRE	(1 << 5)	/* THR empty */

/* this returns the baud divisor for the specified baudrate */
#define BAUD_DIV(x)	(CPU_CLOCK / (16 * x))

/* in isr.S */
extern void isr_uart(void);

void uart_init(void)
{
	/* init FIFO */
	FCR = FIFO_ENABLE | RX_RESET | TX_RESET;

	/* set baudrate */
	LCR = DLAB;
	DLL = BAUD_DIV(CONFIG_UART_BAUD) & 0xff;
	DLM = (BAUD_DIV(CONFIG_UART_BAUD) & 0xff00) >> 8;

	/* unset DLAB and set bits per byte */
	LCR = BPB(8);

	/* enable interrupt on Receive Data Available */
	vic_set(UART_IRQ, isr_uart);
	IER = 1;
}

void uart_putchar(char c)
{
	while (!(LSR & THRE)) ;

	THR = c;
}

static struct term_info *ti;

void uart_late_init(void)
{
	ti = term_init("ttyUART0", uart_putchar);
}

/* called by asm function in isr.S */
extern void uart_irq(struct context *);

void uart_irq(struct context *ctx)
{
	while (LSR & RDR)
		term_newc(ctx, ti, RBR);

	vic_reset();
}
